/*
 *
 * URL:       $URL$
 * Revision:  $Revision: #1 $
 * Date:      $Date: 2017/11/02 $
 * Author:    $Author: build $
 *
 * (C) Copyright 2002-2012 Riverbed Technology, Inc.
 * All rights reserved.
 *
 */

static const char rcsid[] = "$Id: //depot/prod/cag/R4.4.1x/net-snmp/net-snmp-5.7.1/agent/sn_main.c#1 $";

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "common.h"
#include "logging.h"
#include "bnode.h"
#include "sn_main.h"
#include "sn_mgmt.h"
#include "sn_modules.h"
#include "sn_registration.h"
#include "sn_scalars.h"
#include "sn_tables.h"

const char *snmpd_program_name = "snmpd";
tbool sn_busy = false;
extern size_t max_buff_size;

/* ------------------------------------------------------------------------- */
void
sn_main_init(void)
{
    int err = 0;

    err = lc_log_init(snmpd_program_name, "", LCO_none, 
                      LC_COMPONENT_ID(LCI_SNMPD),
                      LOG_DEBUG, LOG_LOCAL0, LCT_SYSLOG);
    bail_error(err);

    /* Load modules first to make debugging easier */
    err = sn_modules_load();
    bail_error(err);

    /* Open session with mgmtd */
    err = sn_mgmt_init();
    bail_error(err);

    /* Get ready for registration */
    err = sn_reg_init();
    bail_error(err);

    /* Let modules register their MIBs with us */
    err = sn_modules_init();
    bail_error(err);

    /* Tell NET-SNMP all of the registration info we have gathered */
    err = sn_reg_init_netsnmp();
    bail_error(err);

 bail:
    return;
}


/* ------------------------------------------------------------------------- */
void
sn_main_deinit(void)
{
    int err = 0;

    /*
     * Not clear whether we could be called at any time (e.g. with
     * outstanding mgmt messages) or only from NET-SNMP's main loop.
     * To be safe, just close GCL sessions, but do not deinit the GCL.
     * Deinitializing the GCL entirely is not really necessary if we
     * are exiting anyway.
     */
    err = sn_mgmt_shutdown();
    bail_error(err);

 bail:
    return;
}


/* ------------------------------------------------------------------------- */
/* The following two functions are the callbacks called by NET-SNMP
 * whenever a variable we have registered is requested.  There are
 * separate functions for scalar and columnar variables because they
 * require separate handling, and because pointers to structures of
 * different types are passed as the magic number.
 *
 * \param vp The variable registration structure provided at the time
 * the MIB was registered with NET-SNMP for the variable expected to
 * be returned.
 *
 * \param name The OID taken directly from the request.  If this is a
 * GET request, it should be a prefix of, or identical to, the OID in
 * \c vp.  If this is a GETNEXT request, this is probably the previous
 * node.
 *
 * \param length Number of elements in the OID in the \c name parameter.
 *
 * \param exact If 0, this is a GETNEXT request; if 1, this is a GET
 * request.
 *
 * \retval name The OID of the variable being returned, to be written over
 * the name requested.
 *
 * \retval length The length of the OID of the variable being
 * returned, to be written over the length of the name requested.
 *
 * \retval var_len The size of the data being returned, in bytes.
 *
 * \retval write_method This does not apply to us; it deals with 
 * SNMP SETs.
 *
 * \return A pointer to the data itself.  Note that the global
 * variable long_return can be used to hold long int return values (up
 * to 32 bits, presumably), and return_buf can be used to hold string
 * return values (up to max_buff_size bytes).
 */


/* ------------------------------------------------------------------------- */
u_char *
sn_get_handler_scalar(struct variable *vp, oid *name, size_t *length,
		      int exact, size_t *var_len,
		      WriteMethod ** write_method)
{
    int err = 0;
    sn_scalar *scalar = NULL;
    oid req_oid[MAX_OID_LEN];
    size_t req_oid_len = 0;		

    bail_null(vp);
    scalar = (sn_scalar *)(vp->magic);
    bail_null(scalar);

    req_oid_len = *length;
    memcpy(req_oid, name, req_oid_len * sizeof(oid));

    /*
     * We should never get called reentrantly because our own inner
     * loop does not listen on the SNMP port on which new GET/GETNEXT
     * requests come.
     */
    bail_require(!sn_busy);
    sn_busy = true;
    
    err = sn_scalar_get
	(scalar, vp->name, vp->namelen, req_oid, req_oid_len, exact,
	 name, length, return_buf, max_buff_size, var_len);
    complain_error(err);

    sn_busy = false;
    *write_method = NULL;

    err = sn_mgmt_handle_queued();
    bail_error(err);

 bail:
    if (err || *var_len == -1) {
	return(NULL);
    }
    else {
	return(return_buf);
    }    
}


/* ------------------------------------------------------------------------- */
u_char *
sn_get_handler_columnar(struct variable *vp, oid *name, size_t *length,
			int exact, size_t *var_len,
			WriteMethod ** write_method)
{
    int err = 0;
    sn_field *field = NULL;
    oid req_oid[MAX_OID_LEN];
    size_t req_oid_len = 0;		

    bail_null(vp);
    field = (sn_field *)(vp->magic);
    bail_null(field);

    req_oid_len = *length;
    memcpy(req_oid, name, req_oid_len * sizeof(oid));

    /*
     * We should never get called reentrantly because our own inner
     * loop does not listen on the SNMP port on which new GET/GETNEXT
     * requests come.
     */

    if (field->sf_mgmt_type == bt_uint64) {
        vp->type = ASN_COUNTER64;
    }

    bail_require(!sn_busy);
    sn_busy = true;

    err = sn_table_get_field
	(field, vp->name, vp->namelen, req_oid, req_oid_len, exact,
	 name, length, return_buf, max_buff_size, var_len);
    complain_error(err);

    sn_busy = false;
    *write_method = NULL;

    err = sn_mgmt_handle_queued();
    bail_error(err);

 bail:
    if (err || *var_len == -1) {
	return(NULL);
    }
    else {
	return(return_buf);
    }    
}
