/*
 *
 * URL:       $URL$
 * Revision:  $Revision: #1 $
 * Date:      $Date: 2017/11/02 $
 * Author:    $Author: build $
 *
 * (C) Copyright 2002-2012 Riverbed Technology, Inc.
 * All rights reserved.
 *
 */

static const char rcsid[] = "$Id: //depot/prod/cag/R4.4.1x/net-snmp/net-snmp-5.7.1/agent/sn_mgmt.c#1 $";

#include <unistd.h>
#include <signal.h>
#include "common.h"
#include "bnode.h"
#include "gcl.h"
#include "fdintrcb.h"
#include "hash_table.h"
#include "sn_main.h"
#include "sn_notifs.h"
#include "md_client.h"
#include "sn_mgmt.h"
#include <net-snmp/library/snmpusm.h>

gcl_handle *sn_gcl_handle = NULL;
gcl_session *sn_mgmt_session = NULL;
fdic_handle *sn_fdich = NULL;
bn_response *sn_mgmt_response = NULL;

array *sn_pending_requests = NULL;

/*
 * This is separate from sn_mgmt_response (we could have just tested
 * it for non-NULL) so we can detect cases where the receiving or
 * copying of the reply fails.
 */
tbool sn_mgmt_awaiting_reply = false;

tbool sn_mgmt_request_cancelled = false;

/*
 * Keys point to dynamically-allocated uint32s.
 * Values are dynamically-allocated (fdic_event *).
 */
ht_table *sn_gcl_events = NULL;

static int sn_mgmt_open_gcl_session(void);

static int sn_event_fdic_callback(const fdic_event_array *session_events,
                                  void *re_arg);

static void sn_gcl_event_readable_handler(int fd, void *data);

static void sn_gcl_event_writable_handler(int fd, void *data);

static int sn_fdic_event_free(void **key, int32 key_len, void **value);

static int sn_mgmt_handle_response(gcl_session *sess,
				   const bn_response *response, void *arg);

static int sn_mgmt_handle_event(gcl_session *sess,
				const bn_request *event_request, void *arg);

int sn_mdc_msg_handler(void *data, gcl_session *session,
		       bn_request *request, bn_response **ret_response);


/* ------------------------------------------------------------------------- */
/* Handle an action request forwarded to us by the mgmt daemon
 */
int
sn_mgmt_handle_action_request(gcl_session *sess,
			      const bn_request *action_request, void *arg);

/* ------------------------------------------------------------------------- */
int
sn_mgmt_init(void)
{
    int err = 0;
    ht_options ht_opts;

    /* .......................................................................
     * First create the hashtable that will hold the events the GCL
     * gives us.
     */
    err = ht_default_options(&ht_opts);
    bail_error(err);

    ht_opts.ho_key_value_dup_func = NULL;
    ht_opts.ho_key_value_free_func = sn_fdic_event_free;
    ht_opts.ho_insert_fails_duplicates = true;

    err = ht_new(&sn_gcl_events, &ht_opts);
    bail_error_null(err, sn_gcl_events);

    /* .......................................................................
     * Now open the GCL session.  The GCL retries for us on failure,
     * so we don't have to.
     */

    err = sn_mgmt_open_gcl_session();
    complain_error(err);

    /* .......................................................................
     * And tell the md_client library about us
     */

    if (err == 0) {
	err = mdc_init(sn_mdc_msg_handler, NULL, NULL, NULL, sn_mgmt_session);
	bail_error(err);
    }

 bail:
    if (err) {
	lc_log_debug(LOG_ERR, "Unable to connect to the management daemon");
	sn_mgmt_session = NULL;
	err = 0;
    }
    else {
	lc_log_basic(LOG_INFO, "Connected to the management daemon");
    }
    return(err);    
}


/* ------------------------------------------------------------------------- */
static int
sn_mgmt_open_gcl_session(void)
{
    int err = 0;
    gcl_global_options gcl_opts;

    err = fdic_init(&sn_fdich, sn_event_fdic_callback, NULL);
    bail_error(err);

    err = gcl_options_get_defaults(&gcl_opts);
    bail_error(err);

    err = gcl_init(&sn_gcl_handle, &gcl_opts, sn_fdich);
    bail_error(err);

    err = bn_set_global_callback_query_response_msg
	(sn_gcl_handle, sn_mgmt_handle_response, NULL);
    bail_error(err);

    err = bn_set_global_callback_action_response_msg
	(sn_gcl_handle, sn_mgmt_handle_response, NULL);
    bail_error(err);

    err = bn_set_global_callback_event_request_msg
	(sn_gcl_handle, sn_mgmt_handle_event, NULL);
    bail_error(err);

    err = bn_set_global_callback_action_request_msg
	(sn_gcl_handle, sn_mgmt_handle_action_request, NULL);
    bail_error(err);

    err = gcl_session_client_new(sn_gcl_handle, gcl_client_snmpd,
				 NULL, 0, &sn_mgmt_session);
    bail_error(err);

    err = gcl_session_client_connect_by_provider
	(sn_mgmt_session, gcl_provider_mgmtd, true);
    if (err == lc_err_io_error || err == lc_err_io_closed) {
        goto bail;
    }
    bail_error(err);

    err = gcl_session_client_open_service_blocking
	(sn_mgmt_session, gbp_proto_bnode, bap_service_mgmt, NULL, NULL, -1);
    bail_error(err);

 bail:
    if (err && sn_mgmt_session) {
	gcl_session_close(sn_mgmt_session, false);
	gcl_session_free(&sn_mgmt_session);
    }
    return(err);
}


/* ------------------------------------------------------------------------- */
int
sn_mgmt_shutdown(void)
{
    int err = 0;

    err = gcl_shutdown(sn_gcl_handle);
    bail_error(err);

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_event_fdic_callback(const fdic_event_array *session_events,
                       void *re_arg)
{
    int err = 0;
    uint32 num_events = 0, i = 0;
    tbool found = 0;
    uint32 ev_id = 0;
    uint32 *new_ev_id = NULL;
    const fdic_event *fe = NULL;
    fdic_event *fe_copy = NULL;

    num_events = fdic_event_array_length_quick(session_events);
    for (i = 0; i < num_events; ++i) {
	fe = fdic_event_array_get_quick(session_events, i);
	bail_null(fe);

	ev_id = fe->fee_event_id;

	err = ht_lookup(sn_gcl_events, &ev_id, sizeof(ev_id), &found,
			NULL, NULL, (void **)&fe_copy);
	bail_error(err);

	/*
	 * If we were asked to delete this event
	 */
	if (fe->fee_active == false) {
	    /*
	     * GCL won't ask us to delete events that weren't
	     * previously registered with us.
	     */
	    bail_require(found);
	    bail_null(fe_copy);
	     
	    if (fe_copy->fee_fd_on_readable) {
		lc_log_debug(LOG_DEBUG,"Don't listen for readability of fd %d",
			     fe_copy->fee_fd);
		unregister_readfd(fe_copy->fee_fd);
	    }
	    if (fe_copy->fee_fd_on_writable) {
		lc_log_debug(LOG_DEBUG,"Don't listen for writability of fd %d",
			     fe_copy->fee_fd);
		unregister_writefd(fe_copy->fee_fd);
	    }
	    err = ht_delete(sn_gcl_events, &ev_id, sizeof(ev_id));
	    bail_error(err);

	    /*
	     * The deletion of this event also heralds that our
	     * session with mgmtd has closed.  Since we haven't
	     * implemented deinitialization yet, this can only mean
	     * that mgmtd has died or broken it off (most likely the
	     * former).
	     *
	     * Our reaction is to cancel any current outstanding mgmt
	     * requests (since we know they will never be answered),
	     * and raise SIGTERM, causing snmpd to exit as soon as
	     * control returns to the main loop.  PM will then
	     * relaunch us.
	     */
	    sn_mgmt_request_cancelled = true;
	    raise(SIGTERM);

	    continue;
	}

	/*
	 * Otherwise, we were asked to create or modify an event
	 * record.  If an existing record wasn't found, create a new
	 * one and toss it in the hash table.  After that, both cases
	 * are the same.
	 */

	if (!found) {
	    fe_copy = malloc(sizeof(fdic_event));
	    bail_null(fe_copy);
	    new_ev_id = malloc(sizeof(ev_id));
	    bail_null(new_ev_id);
	    *new_ev_id = ev_id;
	    err = ht_insert(sn_gcl_events, new_ev_id, sizeof(ev_id),
			    fe_copy);
	    bail_error(err);
	}

	/*
	 * Make a copy of callback function pointers, data, etc.
	 */
	*fe_copy = *fe;

	/*
	 * XXX/EMT: NET-SNMP comes with support for registering FDs but
	 * not for registering timers.  It would be some work to add
	 * it, and the GCL doesn't use timers at the moment, so just
	 * don't support it for now...
	 */
	bail_require(fe->fee_uses_timeout == false);

	/*
	 * Not sure why this would ever be false at this point...
	 */
	if (!(fe->fee_uses_fd)) {
	    lc_log_debug(LOG_WARNING, "Got called with fee_uses_fd set to "
			 "false");
	    continue;
	}

	if (fe->fee_fd_on_readable) {
	    lc_log_debug(LOG_DEBUG, "Registering to hear when fd %d becomes "
			 "readable", fe->fee_fd);
	    err = register_readfd
		(fe->fee_fd, sn_gcl_event_readable_handler, (void *)fe_copy);
	    bail_error(err);
	}

	if (fe->fee_fd_on_writable) {
	    lc_log_debug(LOG_DEBUG, "Registering to hear when fd %d becomes "
			 "writable", fe->fee_fd);
	    err = register_writefd
		(fe->fee_fd, sn_gcl_event_writable_handler,
		 (void *)fe_copy);
	    bail_error(err);
	}
    }

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static void
sn_gcl_event_readable_handler(int fd, void *data)
{
    int err = 0;
    const fdic_event *fe = NULL;

    fe = data;
    bail_null(fe);

    bail_require(fd == fe->fee_fd);
    bail_require(fe->fee_uses_fd && fe->fee_fd_on_readable);

    err = (*(fe->fee_fd_readable_callback_func))
	(fe->fee_fd, fct_read, fe->fee_per_fd_arg,
         fe->fee_fd_readable_callback_arg);
    bail_error(err);

 bail:
    return;
}


/* ------------------------------------------------------------------------- */
static void
sn_gcl_event_writable_handler(int fd, void *data)
{
    int err = 0;
    const fdic_event *fe = NULL;

    fe = data;
    bail_null(fe);

    bail_require(fd == fe->fee_fd);
    bail_require(fe->fee_uses_fd);
    bail_require_quiet(fe->fee_fd_on_writable);

    err = (*(fe->fee_fd_writable_callback_func))
	(fe->fee_fd, fct_write, fe->fee_per_fd_arg,
         fe->fee_fd_writable_callback_arg);
    bail_error(err);

 bail:
    return;
}


/* ------------------------------------------------------------------------- */
static int
sn_fdic_event_free(void **key, int32 key_len, void **value)
{
    int err = 0;
    uint32 **my_key = (uint32 **)key;
    fdic_event **my_value = (fdic_event **)value;

    if (my_key) {
	safe_free(*my_key);
    }

    if (my_value) {
	safe_free(*my_value); /* It was just shallow-copied before */
    }

bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
int
sn_mgmt_send_msg(bn_request *req, bn_response **ret_resp)
{
    int err = 0;
    int numfds = 0, i = 0, count = 0;
    fd_set readfds, writefds;
    lt_time_sec req_sent = 0;
    lt_dur_sec timeout_sec = 0;
    struct timeval timeout_tv = {0, 0};

    bail_null(req);
    bail_null(ret_resp);
    *ret_resp = NULL;

    /*
     * Catch any attempts at reentrancy, as we cannot deal with more
     * than one message outstanding at a time.
     *
     * Reentrancy should also be prevented at a higher level.  There
     * are two things that could trigger us to make management
     * requests: GET requests and management events.  We cannot
     * receive other GET requests while waiting for a management
     * reply, since the SNMP port is not in our list of FDs to watch.
     * If we receive an event while processing a GET request or other
     * event, is is enqueued, and dealt with after the current
     * activity is complete.
     */

    bail_require(sn_mgmt_awaiting_reply == false);
    sn_mgmt_awaiting_reply = true;

    sn_mgmt_response = NULL;
    sn_mgmt_request_cancelled = false;

    err = bn_request_msg_send(sn_mgmt_session, req);
    bail_error(err);

    req_sent = lt_curr_time();

    while (sn_mgmt_awaiting_reply == true &&
	   sn_mgmt_request_cancelled == false) {

	/*
	 * The following was cut-and-pasted from snmpd.c
	 * around lines 1062-1155.  If only it were in functions
	 * that we could call, instead of embedded in their big
	 * receive() function.  We can't return to NET-SNMP's
	 * own main loop because of flow-of-control issues;
	 * the request must look synchronous to the caller.
	 */ 

        numfds = 0;
        FD_ZERO(&readfds);
        FD_ZERO(&writefds);

        for (i = 0; i < external_readfdlen; i++) {
            FD_SET(external_readfd[i], &readfds);
            if (external_readfd[i] >= numfds)
                numfds = external_readfd[i] + 1;
        }

        for (i = 0; i < external_writefdlen; i++) {
            FD_SET(external_writefd[i], &writefds);
            if (external_writefd[i] >= numfds)
                numfds = external_writefd[i] + 1;
        }

	/*
	 * The timeout is relative to when we intially sent the message.
	 * If some other event occurs after we send the message, we 
	 * don't want to postpone the timeout for receiving our reply.
	 */
	timeout_sec =
	    max(0, sn_mgmt_timeout_sec - (lt_curr_time() - req_sent));
	err = lt_time_sec_to_timeval(timeout_sec, &timeout_tv);
	bail_error(err);	

        count = select(numfds, &readfds, &writefds, NULL, &timeout_tv);

        if (count > 0) {
            for (i = 0; count && (i < external_readfdlen); i++) {
                if (FD_ISSET(external_readfd[i], &readfds)) {
                    DEBUGMSGTL(("snmpd/select", "readfd[%d] = %d\n",
                                i, external_readfd[i]));
                    external_readfdfunc[i] (external_readfd[i],
                                            external_readfd_data[i]);
                    FD_CLR(external_readfd[i], &readfds);
                    count--;
                }
            }

            for (i = 0; count && (i < external_writefdlen); i++) {
                if (FD_ISSET(external_writefd[i], &writefds)) {
                    DEBUGMSGTL(("snmpd/select", "writefd[%d] = %d\n",
                                i, external_writefd[i]));
                    external_writefdfunc[i] (external_writefd[i],
                                             external_writefd_data[i]);
                    FD_CLR(external_writefd[i], &writefds);
                    count--;
                }
            }
	}

	else {
	    lc_log_basic(LOG_WARNING, "Management request timed out after "
			 "%d seconds; giving up", sn_mgmt_timeout_sec);
	    break;
	}
    }

    if (sn_mgmt_request_cancelled) {
	err = lc_err_cancelled;
	sn_mgmt_request_cancelled = false;
	lc_log_debug(LOG_NOTICE, "Cancelling outstanding mgmt request");
    }

    *ret_resp = sn_mgmt_response;
    sn_mgmt_response = NULL;
    
 bail:
    sn_mgmt_awaiting_reply = false;
    return(err);
}


/* ------------------------------------------------------------------------- */
int 
sn_mdc_msg_handler(void *data, gcl_session *session,
		   bn_request *request, bn_response **ret_response)
{
    int err = 0;

    err = sn_mgmt_send_msg(request, ret_response);
    bail_error(err);

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_mgmt_handle_response(gcl_session *sess, const bn_response *response,
			void *arg)
{
    int err = 0;

    sn_mgmt_awaiting_reply = false;

    err = bn_response_msg_dup(response, &sn_mgmt_response);
    bail_error(err);

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_mgmt_handle_event(gcl_session *sess, const bn_request *event_request,
		     void *arg)
{
    int err = 0;
    bn_request *req_copy = NULL;
    array_options aopts;

    if (sn_pending_requests == NULL) {
        err = array_options_get_defaults(&aopts);
        bail_error(err);
        aopts.ao_elem_dup_func = array_std_assign_func;
        err = array_new(&sn_pending_requests, &aopts);
        bail_error_null(err, sn_pending_requests);
    }

    if (sn_busy) {
	err = bn_request_msg_dup(event_request, &req_copy);
	bail_error_null(err, req_copy);
	err = array_append(sn_pending_requests, req_copy);
	bail_error(err);
	lc_log_basic(LOG_INFO, "Already busy with something else; "
		     "queueing event to handle later");
	goto bail;
    }

    err = sn_notif_handle_event_request(event_request);
    complain_error(err);

    err = sn_mgmt_handle_queued();
    bail_error(err);

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
int
sn_mgmt_handle_queued(void)
{
    int err = 0;
    bn_request *req_copy = NULL;

    while (array_length_quick(sn_pending_requests) > 0) {
	err = array_remove(sn_pending_requests, 0, (void **)&req_copy);
	bail_error(err);
	lc_log_basic(LOG_INFO, "Now processing queued event");
	err = sn_notif_handle_event_request(req_copy);
	complain_error(err);
	bn_request_msg_free(&req_copy);
    }

 bail:
    return(err);
}

/* ------------------------------------------------------------------------- */
int
sn_mgmt_handle_action_request(gcl_session *sess,
			      const bn_request *action_request, void *arg)
{
    int err = 0, err2 = 0;
    tstring *action_name = NULL;
    bn_response *resp = NULL;
    bn_binding *binding = NULL;
    const bn_binding *password_binding = NULL;
    const bn_binding *hf_binding = NULL;
    bn_binding_array *bindings = NULL;
    bn_msg_type req_type = bbmt_none;
    char *password = NULL, *hash_function = NULL, *auth_key = NULL;

    err = bn_request_get
	(action_request, &req_type, NULL, true, &bindings, NULL, NULL);
    bail_error(err);
    bail_require(req_type == bbmt_action_request);
    
    err = bn_action_request_msg_get_action_name(action_request, &action_name);
    bail_error_null(err, action_name);

    err = bn_action_response_msg_create
	(&resp, bn_request_get_msg_id(action_request), err, NULL);
    complain_error(err);

    if (ts_equal_str(action_name, "/snmp/usm/actions/generate_auth_key", false))
    {
        lc_log_basic(LOG_INFO, "Received action %s", ts_str(action_name));
        err = bn_binding_array_get_binding_by_name(bindings, "password", 
                &password_binding);
        bail_error(err);

        if (password_binding != NULL) {
            err = bn_binding_get_str
                (password_binding, ba_value, bt_string, NULL, &password);
            bail_error(err); 
        }    
 
        err = bn_binding_array_get_binding_by_name(bindings, "hash_function", 
                &hf_binding);
        bail_error(err);

        if (hf_binding != NULL) {
            err = bn_binding_get_str
                (hf_binding, ba_value, bt_string, NULL, &hash_function);
            bail_error(err); 
        }    
     
        err = generate_auth_key(password, hash_function, &auth_key);
        bail_error(err);

        err = bn_binding_new_str(&binding, "auth_key", ba_value, bt_string, 0,
                auth_key);
        bail_error_null(err, binding);

        err = bn_response_msg_append_binding(resp, binding, 1);
        bail_error(err);
    }

 bail:    
     if (resp) {
        err2 = bn_action_response_set_return(resp, err, NULL);
        complain_error(err2);

        err2 = bn_response_msg_send(sess, resp);
        complain_error(err2);
    }
     ts_free(&action_name);
     bn_binding_free(&binding);
     bn_binding_array_free(&bindings);
     safe_free(password);
     safe_free(hash_function);
     safe_free(auth_key);
     bn_response_msg_free(&resp);
     return err;
}
