/*
 *
 * URL:       $URL$
 * Revision:  $Revision: #1 $
 * Date:      $Date: 2017/11/02 $
 * Author:    $Author: build $
 *
 * (C) Copyright 2002-2012 Riverbed Technology, Inc.
 * All rights reserved.
 *
 */

static const char rcsid[] = "$Id: //depot/prod/cag/R4.4.1x/net-snmp/net-snmp-5.7.1/agent/sn_notifs.c#1 $";

#include "common.h"
#include "sn_main.h"
#include "sn_registration.h"
#include "sn_utils.h"
#include "md_client.h"
#include "sn_notifs.h"

/*
 * Needed for netsnmp_get_agent_uptime().
 */
#include "snmp_agent.h"

oid sn_oid_sysuptime[] = {1, 3, 6, 1, 2, 1, 1, 3, 0};
oid sn_oid_snmptrapoid[] = {1, 3, 6, 1, 6, 3, 1, 1, 4, 1, 0};

/* ----------------------------------------------------------------------------
 * Search all of the registered MIBs to find the sn_notif record,
 * if any, that pertains to the specified event.
 */
static int sn_notif_find_record(const char *event_name, sn_notif **ret_notif);


/* ----------------------------------------------------------------------------
 * Send a notification of the specified type, given the bindings that
 * arrived with the event message.
 */
static int sn_notif_convert_and_send(sn_notif *notif,
				     bn_binding_array *bindings);


/* ------------------------------------------------------------------------- */
static int sn_notif_get_event_vars(sn_notif *notif, bn_binding_array *bindings,
				   netsnmp_variable_list **ret_vars);


/* ------------------------------------------------------------------------- */
static int sn_notif_get_event_var(sn_notif_var *var_reg,
				  bn_binding_array *bindings,
				  netsnmp_variable_list *ret_var);


/* ------------------------------------------------------------------------- */
static int sn_notif_get_event_var_scalar(sn_notif_var *var_reg,
					 bn_binding_array *bindings,
					 const char **ret_binding_name,
					 oid *ret_oid_name,
					 uint8 *ret_oid_name_len);


/* ------------------------------------------------------------------------- */
static int sn_notif_get_event_var_field(sn_notif_var *var_reg,
					bn_binding_array *bindings,
					const char **ret_binding_name,
					oid *ret_oid_name,
					uint8 *ret_oid_name_len);

/* ------------------------------------------------------------------------- */
static int sn_notif_get_row_root_name(sn_field *field,
				      bn_binding_array *bindings,
				      char **row_root_name);


/* ------------------------------------------------------------------------- */
static int sn_notif_get_row_root_name_one(const bn_binding_array *bindings,
					  uint32 idx,
					  const bn_binding *binding,
					  const tstring *name,
					  const tstr_array *name_components,
					  const tstring *name_last_part,
					  const tstring *value,
					  void *callback_data);


/* ------------------------------------------------------------------------- */
static int sn_notif_get_var_oid(sn_field *field, const char *row_root_name,
				oid *ret_oid_name, uint8 *ret_oid_name_len);


/* ------------------------------------------------------------------------- */
static int sn_notif_get_binding_name(sn_field *field,
				     bn_binding_array *bindings,
				     const char *row_root_name,
				     const char **ret_binding_name);


/* ------------------------------------------------------------------------- */
typedef struct sn_notif_context {
    char **snc_name_ptr;
    uint16 snc_name_num_parts;
} sn_notif_context;


/* ------------------------------------------------------------------------- */
/** Construct and send an SNMP trap given the set of inputs necessary
 * to send a v2 trap.  Takes the OID of a trap to be sent and a list
 * of variables to be sent with the trap.  Prepends the two mandatory
 * variables for each trap: values for sysUpTime.0 (by calling a
 * NET-SNMP API function, chosen over querying mgmtd in order to keep
 * it in sync with the real sysUpTime.0 variable) and snmpTrapOID.0
 * (using the trap OID passed).  The trap is then sent to all
 * configured sinks.  v2 sinks get the trap as constructed.  v1 sinks
 * get the trap converted into a v1 trap; an enterprise-specific trap
 * is sent, with the specific trap number being the last component of
 * the trap OID provided.
 *
 * \param trap_oid OID of the trap to send.
 *
 * \param trap_oid_len Number of components in the trap OID provided.
 * (note: NOT the size of the OID in bytes)
 *
 * \param vars Linked list of variables to be sent with this trap.
 * If no additional variables are to be sent, NULL may be passed.
 */
static int sn_notif_send_trap_v2(oid *trap_oid, size_t trap_oid_len, 
				 netsnmp_variable_list *vars);

/* ------------------------------------------------------------------------- */
int
sn_notif_handle_event_request(const bn_request *event_request)
{
    int err = 0;
    bn_msg_type msg_type = bbmt_none;
    bn_binding_array *bindings = NULL;
    tstring *event_name = NULL;

    bail_require(!sn_busy);
    sn_busy = true;

    err = bn_request_get(event_request, &msg_type, NULL, false, &bindings,
			 NULL, NULL);
    bail_error(err);
    bail_require(msg_type == bbmt_event_request);

    err = bn_event_request_msg_get_event_name(event_request, &event_name);
    bail_error_null(err, event_name);

    lc_log_basic(LOG_DEBUG, "Handling event %s", ts_str(event_name));

    err = sn_notif_handle_event(ts_str(event_name), bindings);
    bail_error(err);

 bail:
    sn_busy = false;
    ts_free(&event_name);
    bn_binding_array_free(&bindings);
    return(err);
}

/* ------------------------------------------------------------------------- */
int
sn_notif_handle_event(const char *event_name, bn_binding_array *bindings)
{
    int err = 0;
    sn_notif *notif = NULL;

    err = sn_notif_find_record(event_name, &notif);
    bail_error(err);

    if (notif) {
	err = sn_notif_convert_and_send(notif, bindings);
	bail_error(err);
    }

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_notif_find_record(const char *event_name, sn_notif **ret_notif)
{
    int err = 0;
    uint32 num_mibs = 0, i = 0;
    sn_mib *mib = NULL;
    tbool found = false;

    bail_null(event_name);
    bail_null(ret_notif);
    *ret_notif = NULL;

    num_mibs = array_length_quick(sn_reg_mibs);
    for (i = 0; i < num_mibs; ++i) {
	mib = array_get_quick(sn_reg_mibs, i);
	bail_null(mib);

	err = ht_lookup(mib->sm_notifs, event_name, strlen(event_name),
			&found, NULL, NULL, (void **)ret_notif);
	bail_error(err);

	if (found) {
	    break;
	}
    }

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_notif_convert_and_send(sn_notif *notif, bn_binding_array *bindings)
{
    int err = 0;
    netsnmp_variable_list *vars = NULL;
    oid trap_oid[MAX_OID_LEN];
    uint8 trap_oid_len = 0;

    err = sn_notif_get_event_vars(notif, bindings, &vars);
    complain_error(err);
    err = 0;

    switch (notif->sno_notif_type) {

    case snt_trap_v1:
	bail_force(lc_err_nyi);
	break;

    case snt_trap_v2c:
	err = sn_oid_concat(trap_oid, MAX_OID_LEN, &trap_oid_len, 2,
			    notif->sno_mib->sm_base_oid,
			    notif->sno_mib->sm_base_oid_len,
			    notif->sno_trap_oid, notif->sno_trap_oid_len);
	bail_error(err);
	err = sn_notif_send_trap_v2(trap_oid, trap_oid_len, vars);
	bail_error(err);
	break;
	
    default:
	bail_force(lc_err_unexpected_case);
	break;
    }

 bail:
    sn_free_var_list(&vars);
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_notif_get_event_vars(sn_notif *notif, bn_binding_array *bindings,
			netsnmp_variable_list **ret_vars)
{
    int err = 0;
    netsnmp_variable_list *var = NULL, *prev_var = NULL;
    uint32 i = 0;

    bail_null(notif);
    bail_null(bindings);
    bail_null(ret_vars);
    *ret_vars = NULL;

    for (i = 0; i < notif->sno_num_vars; ++i) {
	var = (netsnmp_variable_list *) malloc(sizeof(netsnmp_variable_list));
	bail_null(var);
	if (i == 0) {
	    *ret_vars = var;
	}
	else {
	    bail_null(prev_var);
	    prev_var->next_variable = var;
	}
	memset(var, 0, sizeof(netsnmp_variable_list));
	err = sn_notif_get_event_var
	    (&(notif->sno_vars[i]), bindings, var);
	bail_error(err);
	prev_var = var;
    }

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_notif_get_event_var(sn_notif_var *var_reg,
		       bn_binding_array *bindings,
		       netsnmp_variable_list *ret_var)
{
    int err = 0;
    const char *binding_name = NULL;
    const bn_binding *binding = NULL;
    const char *name = NULL;
    char *val_string = NULL;

    ret_var->name = (oid *) malloc(sizeof(oid) * MAX_OID_LEN);
    bail_null(ret_var->name);

    switch (var_reg->snv_type) {
	
    case snvt_scalar_reg:
	name = var_reg->snv_scalar->ss_snmp_var_name;
	err = sn_notif_get_event_var_scalar
	    (var_reg, bindings, &binding_name, ret_var->name,
	     (uint8 *)&(ret_var->name_length));
	bail_error_msg(err, "Error getting scalar %s for notification", name);
	break;

    case snvt_field_reg:
	name = var_reg->snv_field->sf_snmp_field_name;
	err = sn_notif_get_event_var_field
	    (var_reg, bindings, &binding_name, ret_var->name,
	     (uint8 *)&(ret_var->name_length));
	bail_error_msg(err, "Error getting field %s for notification", name);
	break;

    default:
	bail_force(lc_err_unexpected_case);
	break;
    }

    val_string = calloc(1, sn_return_buf_size);
    bail_null(val_string);

    err = bn_binding_array_get_binding_by_name
	(bindings, binding_name, &binding);
    bail_error_msg(err, "Error getting variable %s for notification", name);

    err = sn_binding_to_snmp_result
	(binding, bt_NONE, (void *)val_string, 256,
	 &(ret_var->val_len), &(ret_var->type));
    bail_error_msg(err, "Error getting variable %s for notification", name);

    snmp_set_var_typed_value(ret_var, ret_var->type, 
            val_string, ret_var->val_len);
	
 bail:
    safe_free(val_string);
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_notif_get_event_var_scalar(sn_notif_var *var_reg,
			      bn_binding_array *bindings,
			      const char **ret_binding_name,
			      oid *ret_oid_name, uint8 *ret_oid_name_len)
{
    int err = 0;
    sn_scalar *scalar = NULL;
    
    scalar = var_reg->snv_scalar;

    *ret_binding_name = scalar->ss_mgmt_node_name;
    err = sn_oid_concat
	(ret_oid_name, MAX_OID_LEN, (uint8 *)ret_oid_name_len, 2,
	 scalar->ss_mib->sm_base_oid, scalar->ss_mib->sm_base_oid_len,
	 scalar->ss_rel_oid, scalar->ss_rel_oid_len);
    bail_error(err);

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_notif_get_event_var_field(sn_notif_var *var_reg,
			     bn_binding_array *bindings,
			     const char **ret_binding_name,
			     oid *ret_oid_name, uint8 *ret_oid_name_len)
{
    int err = 0;
    sn_field *field = NULL;
    char *row_root_name = NULL;

    field = var_reg->snv_field;

    err = sn_notif_get_row_root_name(field, bindings, &row_root_name);
    bail_error_null(err, row_root_name);

    lc_log_debug(LOG_DEBUG, "Got row root name %s", row_root_name);

    err = sn_notif_get_var_oid(field, row_root_name, ret_oid_name,
			       ret_oid_name_len);
    bail_error(err);

    err = sn_notif_get_binding_name(field, bindings, row_root_name,
				    ret_binding_name);
    bail_error(err);

 bail:
    safe_free(row_root_name);
    return(err);
}


/* ------------------------------------------------------------------------- */
/*
  1. Find the name of the binding that is the base of the table row.
  If there is a binding in the event that is a table element, the
  name of the base must be this binding, or one of its parents.
     (a) find the index field
     (b) construct a binding name pattern that is sf_wildcard_parent
         plus a "/ **".
     (c) find a binding that matches this pattern.
     (d) chop off the name after n+1 parts where n is the number of
         parts in sf_wildcard_parent.  Note that the actual
         root binding may not exist in the event, as it may only
         be a child of it.
*/
static int
sn_notif_get_row_root_name(sn_field *field, bn_binding_array *bindings,
			   char **row_root_name)
{
    int err = 0;
    sn_table *table = NULL;
    sn_field *index = NULL;
    char *pattern = NULL;
    uint32 num_matches = 0;
    sn_notif_context ctxt;

    bail_null(row_root_name);
    *row_root_name = NULL;

    table = field->sf_table;

    /* XXX/EMT: will have to change if we support >1 index */
    index = array_get_quick(table->st_fields, 0);
    bail_null(index);
    bail_require(index->sf_index);

    /*
     * XXX/EMT: is it possible that there will be >1 match for this, and
     * that those matches will go to different wildcards?  Is this an
     * important case to support?  Can we just circumvent it by
     * letting the "module" provide a callback function?
     */
    pattern = smprintf("%s/**", index->sf_wildcard_parent);
    bail_null(pattern);

    ctxt.snc_name_ptr = row_root_name;
    ctxt.snc_name_num_parts = index->sf_wildcard_name_index + 1;

    err = mdc_foreach_binding_prequeried
	(bindings, pattern, NULL, sn_notif_get_row_root_name_one, &ctxt,
	 &num_matches);
    bail_error(err);
    bail_require_msg(num_matches > 0, "Couldn't find any matches for %s "
		     "in bindings provided in event", pattern);
    bail_null(*row_root_name);

 bail:
    safe_free(pattern);
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_notif_get_row_root_name_one(const bn_binding_array *bindings,
			       uint32 idx, const bn_binding *binding,
			       const tstring *name,
			       const tstr_array *name_components,
			       const tstring *name_last_part,
			       const tstring *value,
			       void *callback_data)
{
    int err = 0;
    sn_notif_context *ctxt = callback_data;
    tstr_array *parts = NULL;
    uint32 num_elems = 0, i = 0;

    bail_null(ctxt);

    /* If we already have one from a previous binding match, we're done */
    if (*(ctxt->snc_name_ptr)) {
	goto bail;
    }

    err = tstr_array_duplicate(name_components, &parts);
    bail_error_null(err, parts);
    
    num_elems = tstr_array_length_quick(parts);
    for (i = 0; i < num_elems - ctxt->snc_name_num_parts; ++i) {
	err = tstr_array_delete(parts, ctxt->snc_name_num_parts);
	bail_error(err);
    }

    err = bn_binding_name_parts_to_name(parts, true, ctxt->snc_name_ptr);
    bail_error_null(err, *(ctxt->snc_name_ptr));

 bail:
    tstr_array_free(&parts);
    return(err);
}


/* ------------------------------------------------------------------------- */
/*
  2. Find the OID of the variable to include in the trap.
     (a) find the base OID of the field: concat the OIDs of the mib, the
         table, and the field.
     (b) find the value of the index, and append it.
           - if the index is a synthetic sequence number, query the
             children of the index's sf_wildcard_parent, and find
             the name of the root binding among them; the answer is its
             sequence number among them.
*/
static int
sn_notif_get_var_oid(sn_field *field, const char *row_root_name,
		     oid *ret_oid_name, uint8 *ret_oid_name_len)
{
    int err = 0;
    sn_table *table = NULL;
    sn_mib *mib = NULL;
    sn_field *index = NULL; /* XXX/EMT: only supporting one index */
    int32 seq = 0;
    tstr_array *parts = NULL;
    const char *idx_str = NULL;

    bail_null(field);
    table = field->sf_table;
    bail_null(table);
    mib = table->st_mib;
    bail_null(mib);

    err = sn_oid_concat(ret_oid_name, MAX_OID_LEN, ret_oid_name_len, 3,
			mib->sm_base_oid, mib->sm_base_oid_len,
			table->st_rel_oid, table->st_rel_oid_len,
			field->sf_rel_oid, field->sf_rel_oid_len);
    bail_error(err);

    index = array_get_quick(table->st_fields, 0);
    bail_null(index);
    bail_require(index->sf_index);

    switch (index->sf_source) {

    case sfs_child_value:
	bail_force(lc_err_nyi);

    case sfs_wildcard_value:
	err = bn_binding_name_to_parts(row_root_name, &parts, NULL);
	bail_error_null(err, parts);
	idx_str = tstr_array_get_str_quick
	    (parts, tstr_array_length_quick(parts) - 1);
	bail_null(idx_str);
	seq = atol(idx_str);
	bail_require(seq >= 0);
	ret_oid_name[(*ret_oid_name_len)++] = seq;
	lc_log_debug(LOG_DEBUG, "OID index is %d", seq + 1);
	break;

    case sfs_wildcard_sequence:
    break;

    default:
	bail_force(lc_err_unexpected_case);
    }

 bail:
    tstr_array_free(&parts);
    return(err);
}


/* ------------------------------------------------------------------------- */
/*
  3. Find the binding name.  It must be equal to or a child of the 
  table row root name.  Look at the field type. 
     - If it's sfs_wildcard_value, the name is the table root.
       (Note: this would not necessarily be true if there were >1 index)
     - If it's sfs_child_value, it's just the table root plus the
       child name.
     - If it's sfs_wildcard_sequence, THIS IS NOT SUPPORTED,
       because the value does not come from a binding at all.
       That's okay for now, since this would mean it's an index, and
       you're generally not supposed to send indices as trap variables,
       as it is redundant with other info.
*/
static int
sn_notif_get_binding_name(sn_field *field, bn_binding_array *bindings,
			  const char *row_root_name,
			  const char **ret_binding_name)
{
    int err = 0;

    /* XXX/EMT: note: we're returning dyn-allocated name here; caller expects
     * not to have to free -- fix it.
     */

    switch (field->sf_source) {

    case sfs_wildcard_value:
	*ret_binding_name = strdup(row_root_name);
	bail_null(*ret_binding_name);
	break;
	
    case sfs_child_value:
	*ret_binding_name = smprintf("%s/%s", row_root_name, 
				     field->sf_mgmt_node_child_name);
	bail_null(*ret_binding_name);
	break;

    case sfs_wildcard_sequence:
	bail_force(lc_err_nyi);
	
    default:
	bail_force(lc_err_unexpected_case);
    }

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_notif_send_trap_v2(oid *trap_oid, size_t trap_oid_len, 
		      netsnmp_variable_list *vars)
{
    int err = 0;
    netsnmp_variable_list uptime_var;
    netsnmp_variable_list trapoid_var;
    long uptime_hs = 0;        /* uptime in hundreths of a second */
    uint32 code = 0;
    bn_binding *uptime_binding = NULL;

    bail_null(trap_oid);
    bail_require(trap_oid_len > 2); /* Arbitrary minimum */

    uptime_hs = netsnmp_get_agent_uptime();
    uptime_var.next_variable = &trapoid_var;
    uptime_var.name = sn_oid_sysuptime;
    uptime_var.name_length = OID_LENGTH(sn_oid_sysuptime);
    uptime_var.type = ASN_TIMETICKS;
    uptime_var.val.integer = &uptime_hs;
    uptime_var.val_len = sizeof(uptime_hs);
    
    trapoid_var.next_variable = vars;
    trapoid_var.name = sn_oid_snmptrapoid;
    trapoid_var.name_length = OID_LENGTH(sn_oid_snmptrapoid);
    trapoid_var.type = ASN_OBJECT_ID;
    trapoid_var.val.objid = trap_oid;
    trapoid_var.val_len = trap_oid_len * sizeof(oid);

    lc_log_basic(LOG_DEBUG, "Sending trap");
    
    send_v2trap(&uptime_var);

 bail:
    bn_binding_free(&uptime_binding);
    return(err);
}
