/*
 *
 * URL:       $URL$
 * Revision:  $Revision: #1 $
 * Date:      $Date: 2017/11/02 $
 * Author:    $Author: build $
 *
 * (C) Copyright 2002-2012 Riverbed Technology, Inc.
 * All rights reserved.
 *
 */

static const char rcsid[] = "$Id: //depot/prod/cag/R4.4.1x/net-snmp/net-snmp-5.7.1/agent/sn_registration.c#1 $";

#include "common.h"
#include "array.h"
#include "sn_main.h"
#include "sn_registration.h"
#include "sn_utils.h"
#include "md_client.h"

extern int      netsnmp_running;

array *sn_reg_mibs = NULL;

static void sn_mib_free(void *elem);
static void sn_scalar_free(void *elem);
static void sn_table_free(void *elem);
static void sn_field_free(void *elem);
static int sn_notif_free(void **key, int32 key_len, void **value);


/* ************************************************************************* */
/* ************************************************************************* */
/* *** Module interface                                                  *** */
/* ************************************************************************* */
/* ************************************************************************* */


/* ------------------------------------------------------------------------- */
int
sn_reg_init(void)
{
    int err = 0;
    array_options opts;

    err = array_options_get_defaults(&opts);
    bail_error(err);

    opts.ao_elem_free_func = sn_mib_free;

    err = array_new(&sn_reg_mibs, &opts);
    bail_error(err);

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
int
sn_reg_mib_new(sn_mib **ret_mib, const char *mib_name,
	       oid mib_base_oid[], uint8 mib_base_oid_len)
{
    int err = 0;
    sn_mib *new_mib = NULL;
    array_options arr_opts;
    ht_options ht_opts;
    
    bail_null(ret_mib);
    bail_null(mib_name);
    bail_null(mib_base_oid);
    
    new_mib = (sn_mib *) malloc(sizeof(sn_mib));
    bail_null(new_mib);
    
    new_mib->sm_name = mib_name;
    new_mib->sm_base_oid = mib_base_oid;
    new_mib->sm_base_oid_len = mib_base_oid_len;

    err = array_options_get_defaults(&arr_opts);
    bail_error(err);
    arr_opts.ao_elem_free_func = sn_scalar_free;
    err = array_new(&(new_mib->sm_scalars), &arr_opts);
    bail_error(err);

    err = array_options_get_defaults(&arr_opts);
    bail_error(err);
    arr_opts.ao_elem_free_func = sn_table_free;
    err = array_new(&(new_mib->sm_tables), &arr_opts);
    bail_error(err);

    err = ht_default_options(&ht_opts);
    bail_error(err);
    ht_opts.ho_key_equal_func = ht_key_equal_string_func;
    ht_opts.ho_key_value_dup_func = NULL;
    ht_opts.ho_key_value_free_func = sn_notif_free;
    ht_opts.ho_insert_fails_duplicates = true;
    err = ht_new(&(new_mib->sm_notifs), &ht_opts);
    bail_error(err);

    /* The caller should be prepared for the fact that the array owns
     * this memory.
     */
    *ret_mib = new_mib;

    err = array_append_takeover(sn_reg_mibs, (void **) &new_mib);
    bail_error(err);



 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
int
sn_reg_scalar_new(sn_scalar **ret_scalar)
{
    int err = 0;

    bail_null(ret_scalar);

    *ret_scalar = (sn_scalar *) malloc(sizeof(sn_scalar));
    bail_null(*ret_scalar);

    memset(*ret_scalar, 0, sizeof(sn_scalar));
    
 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
int
sn_reg_scalar_register(sn_mib *mib, sn_scalar *scalar)
{
    int err = 0;
    sn_scalar *chk_scalar = NULL;
    uint32 i = 0, num_scalars = 0;

    bail_null(mib);
    bail_null(scalar);

    /*
     * XXX/EMT: validate that the mgmt type and the SNMP type indicated
     * match each other, and that all of the other fields are filled in.
     */

    /*
     * One validation check that is important is to make sure this
     * scalar's oid isn't a duplicate.
     */
    num_scalars = array_length_quick(mib->sm_scalars);
    for (i = 0; i < num_scalars; i++) {
	chk_scalar = (sn_scalar *)array_get_quick(mib->sm_scalars, i);
	bail_null(chk_scalar);

	if (scalar->ss_rel_oid == chk_scalar->ss_rel_oid) {
	    lc_log_debug(LOG_ERR, "Duplicate oid detected (%s and %s)",
			 scalar->ss_snmp_var_name,
			 chk_scalar->ss_snmp_var_name);
	    bail_force_quiet(lc_err_generic_error);
	}
    }

    scalar->ss_mib = mib;

    err = array_append_takeover(mib->sm_scalars, (void **) &scalar);
    bail_error(err);

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
int
sn_reg_table_new(sn_table **ret_table)
{
    int err = 0;
    array_options opts;

    bail_null(ret_table);

    *ret_table = (sn_table *) malloc(sizeof(sn_table));
    bail_null(*ret_table);

    memset(*ret_table, 0, sizeof(sn_table));

    err = array_options_get_defaults(&opts);
    bail_error(err);

    opts.ao_elem_free_func = sn_field_free;

    err = array_new(&((*ret_table)->st_fields), &opts);
    bail_error(err);
    
 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
int
sn_reg_table_register(sn_mib *mib, sn_table *table)
{
    int err = 0;

    bail_null(mib);
    bail_null(table);

    table->st_mib = mib;

    err = array_append_takeover(mib->sm_tables, (void **) &table);
    bail_error(err);

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
int
sn_reg_field_new(sn_field **ret_field)
{
    int err = 0;

    bail_null(ret_field);

    *ret_field = (sn_field *) malloc(sizeof(sn_field));
    bail_null(*ret_field);

    memset(*ret_field, 0, sizeof(sn_field));

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
int
sn_reg_field_register(sn_table *table, sn_field *field)
{
    int err = 0;
    tstr_array *name_parts = NULL;
    const char *part = NULL;
    uint32 num_parts = 0, i = 0, wildcard_index = 0;
    tbool found = false;

    bail_null(table);
    bail_null(field);

    field->sf_table = table;

    if (field->sf_index) {

	table->st_num_indices += 1;
    }

        err = bn_binding_name_to_parts
            (table->st_mgmt_root_pattern, &name_parts, NULL);
        bail_error(err);

        num_parts = tstr_array_length_quick(name_parts);
        wildcard_index = 0;
        found = false;
        for (i = 0; i < num_parts; ++i) {
            part = tstr_array_get_str_quick(name_parts, i);
            bail_null(part);
            if (!strcmp(part, "*")) {
                ++wildcard_index;
                if (wildcard_index == field->sf_wildcard_index) {
                    found = true;
                    break;
                }
            }
        }

        if (!found && field->sf_wildcard_index) {
            lc_log_debug(LOG_ERR, "Field %s in table %s refers to nonexistant "
                    "wildcard #%d in name pattern %s",
                    field->sf_snmp_field_name, table->st_snmp_table_name,
                    field->sf_wildcard_index,
                    table->st_mgmt_root_pattern);
            bail_force_quiet(lc_err_generic_error);
        }

        field->sf_wildcard_name_index = i;

	/*
	 * To make the parent's name, delete this wildcard and all of
	 * the parts after it, then synthesize it all back into a
	 * name.
	 */
	while (i < num_parts) {
        err = tstr_array_delete(name_parts, field->sf_wildcard_name_index);
        bail_error(err);
	    ++i;
	}

	err = bn_binding_name_parts_to_name
	    (name_parts, true, &(field->sf_wildcard_parent));
	bail_error(err);

    /*
       Some of the "single index only" code assumes that the index field is the
       first field. This should not be a problem once we move all the 
       fucntionality to the more generic "composite" index routines.
    */

	err = array_append_takeover(table->st_fields, (void **) &field);
	bail_error(err);

 bail:
    tstr_array_free(&name_parts);
    return(err);
}


/* ------------------------------------------------------------------------- */
int
sn_reg_notif_new(sn_notif **ret_notif, uint16 num_vars)
{
    int err = 0;
    sn_notif *notif = NULL;

    bail_null(ret_notif);
    *ret_notif = NULL;

    notif = (sn_notif *) malloc(sizeof(sn_notif));
    bail_null(notif);
    memset(notif, 0, sizeof(sn_notif));

    if (num_vars) {
	notif->sno_vars = (sn_notif_var *) malloc
	    (num_vars * sizeof(sn_notif_var));
	bail_null(notif->sno_vars);
    }

    notif->sno_num_vars = num_vars;

    *ret_notif = notif;

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
int
sn_reg_notif_register(sn_mib *mib, sn_notif *notif)
{
    int err = 0;
    tstring *alarm_name = NULL;
    bn_binding_array *rl_bindings = NULL;
    bn_binding *enabled_binding = NULL;
    bn_binding *event_binding = NULL;
    bn_binding_array *results = NULL;
    char *rl_config_path = NULL;
    bn_binding *rl_config_path_binding = NULL;

    bail_null(mib);
    bail_null(notif);

    if (netsnmp_running) {
        bail_require_msg(notif->sno_notif_type == snt_trap_v2c, 
                "Only v2 trap registrations are supported currently "
                "(event %s)", notif->sno_event_name);

        notif->sno_mib = mib;

        err = ht_insert(mib->sm_notifs, (void *)(notif->sno_event_name),
                strlen(notif->sno_event_name), notif);
        bail_error(err);

        /* Event interests for alarms need to be rate-limited as per
           user-config.
        */           
        if (strstr(notif->sno_event_name, "/alarm/event/alarm")) {
            err = bn_binding_array_new(&rl_bindings);
            bail_error_null(err, rl_bindings);

            err = bn_binding_name_to_parts_va(notif->sno_event_name,
                    true, 1, 3, &alarm_name, NULL);
            bail_error_null(err, alarm_name);

            err = bn_binding_new_str(&event_binding, "event_name",
                    ba_value, bt_name, 0, notif->sno_event_name);
            bail_error(err);

            err = bn_binding_array_insert(rl_bindings, 0, event_binding);
            bail_error(err);

            err = bn_binding_new_tbool(&enabled_binding, "rl_enabled",
                    ba_value, 0, true);
            bail_error(err);

            err = bn_binding_array_insert(rl_bindings, 1, enabled_binding);
            bail_error(err);

            safe_free(rl_config_path);
            rl_config_path = smprintf("/alarm/config/alarm/%s/handlers/snmp",
                    ts_str(alarm_name));
            bail_null(rl_config_path);

            err = bn_binding_new_str(&rl_config_path_binding, "rl_config_path",
                    ba_value, bt_string, 0, rl_config_path);
            bail_error(err);

            err = bn_binding_array_insert(rl_bindings, 2, 
                    rl_config_path_binding);
            bail_error(err);

            err = mdc_send_action_with_bindings_and_results
                (NULL, NULL, "/mgmtd/events/actions/interest/add",
                 rl_bindings, &results);
            bail_error(err);
        }
        else {
            err = mdc_send_action_with_bindings_str_va
                (NULL, NULL, "/mgmtd/events/actions/interest/add",
                 "event_name", bt_name, notif->sno_event_name, NULL);
            bail_error(err);
        }
    }

 bail:
    ts_free(&alarm_name);
    safe_free(rl_config_path);
    bn_binding_array_free(&rl_bindings);
    bn_binding_array_free(&results);
    bn_binding_free(&enabled_binding);
    bn_binding_free(&event_binding);
    bn_binding_free(&rl_config_path_binding);
    return(err);
}


/* ************************************************************************* */
/* ************************************************************************* */
/* *** Infrastructure interface                                          *** */
/* ************************************************************************* */
/* ************************************************************************* */


/*
 * XXX/EMT: this could be more generalized, but should be sufficient for v1.0.
 */
#define VARSTRUCT            struct variable13
#define VARSTRUCT_oid_size   13

static int sn_reg_get_variable_array(const sn_mib *mib_reg,
				     VARSTRUCT **ret_vars,
				     uint32 *ret_num_vars);

static int sn_reg_get_num_field_regs(const sn_mib *mib_reg,
				     uint32 *ret_num_field_regs);

static int sn_reg_scalar_regs_to_structs(const sn_mib *mib_reg,
					 VARSTRUCT *vars, 
					 uint32 offset, uint32 *ret_offset);

static int sn_reg_table_reg_to_structs(const sn_table *table_reg,
				       VARSTRUCT *vars, uint32 offset,
				       uint32 *ret_offset);

/* ------------------------------------------------------------------------- */
int
sn_reg_init_netsnmp(void)
{
    int err = 0;
    sn_mib *mib_reg = NULL;
    VARSTRUCT *vars = NULL;
    uint32 num_mibs = 0, i = 0, num_vars = 0;

    num_mibs = array_length_quick(sn_reg_mibs);
    for (i = 0; i < num_mibs; ++i) {
	mib_reg = (sn_mib *)array_get_quick(sn_reg_mibs, i);
	bail_null(mib_reg);

	err = sn_reg_get_variable_array(mib_reg, &vars, &num_vars);
	bail_error_null(err, vars);

	err = register_mib(mib_reg->sm_name, (struct variable *)vars,
			   sizeof(VARSTRUCT), num_vars, mib_reg->sm_base_oid,
			   mib_reg->sm_base_oid_len);
	bail_error(err);
    }

    /*
     * XXX/EMT: not sure how vars ever gets deallocated.  NET-SNMP doesn't
     * seem to have a deinitialization story, so maybe it never does.
     * As long as it never gets reinitialized, there's no problem.
     * Should check to see if we get called again on SIGHUP.
     */

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_reg_get_variable_array(const sn_mib *mib_reg, VARSTRUCT **ret_vars,
			  uint32 *ret_num_vars)
{
    int err = 0;
    VARSTRUCT *vars = NULL;
    uint32 num_scalar_regs = 0, num_field_regs = 0, num_vars = 0;
    uint32 num_tables = 0, i = 0, offset = 0;
    sn_table *table_reg = NULL;

    bail_null(mib_reg);
    bail_null(ret_vars);
    bail_null(ret_num_vars);
    *ret_vars = NULL;
    *ret_num_vars = 0;

    num_scalar_regs = array_length_quick(mib_reg->sm_scalars);
    err = sn_reg_get_num_field_regs(mib_reg, &num_field_regs);
    bail_error(err);
    num_vars = num_scalar_regs + num_field_regs;

    /*
     * Technically if num_vars is 0 we should set vars to NULL.
     * But allocate a slot anyway just as a precaution against
     * anyone down the line assuming there's a pointer here.
     */
    vars = (VARSTRUCT *) malloc(sizeof(VARSTRUCT) * max(1, num_vars));
    bail_null(vars);

    offset = 0; /* Array offset for new var structs */

    err = sn_reg_scalar_regs_to_structs(mib_reg, vars, offset, &offset);
    bail_error(err);

    num_tables = array_length_quick(mib_reg->sm_tables);
    for (i = 0; i < num_tables; ++i) {
	table_reg = array_get_quick(mib_reg->sm_tables, i);
	bail_null(table_reg);
	err = sn_reg_table_reg_to_structs(table_reg, vars, offset, &offset);
	bail_error(err);
    }

    *ret_vars = vars;
    *ret_num_vars = num_vars;

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_reg_get_num_field_regs(const sn_mib *mib_reg, uint32 *ret_num_field_regs)
{
    int err = 0;
    uint32 num_tables = 0, i = 0, num_field_regs = 0;
    sn_table *table_reg = NULL;

    bail_null(mib_reg);
    bail_null(ret_num_field_regs);
    *ret_num_field_regs = 0;

    num_tables = array_length_quick(mib_reg->sm_tables);
    for (i = 0; i < num_tables; ++i) {
	table_reg = array_get_quick(mib_reg->sm_tables, i);
	bail_null(table_reg);
	err = array_length(table_reg->st_fields, &num_field_regs);
	bail_error(err);
	*ret_num_field_regs += num_field_regs;
    }

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_reg_scalar_regs_to_structs(const sn_mib *mib_reg, VARSTRUCT *vars, 
			      uint32 offset, uint32 *ret_offset)
{
    int err = 0;
    uint32 num_scalar_regs = 0, i = 0;
    sn_scalar *scalar_reg = NULL;

    bail_null(mib_reg);
    bail_null(vars);
    bail_null(ret_offset);

    num_scalar_regs = array_length_quick(mib_reg->sm_scalars);
    for (i = offset; i < offset + num_scalar_regs; ++i) {
	scalar_reg = array_get_quick(mib_reg->sm_scalars, i - offset);
	bail_null(scalar_reg);

	/* XXX64 should really be 'int_ptr' instead of 'long' */
	bail_require(sizeof(void *) <= sizeof(long));
	vars[i].magic = (long)scalar_reg;

	vars[i].type = scalar_reg->ss_snmp_var_type;
	vars[i].acl = RONLY;
	vars[i].findVar = sn_get_handler_scalar;

	vars[i].namelen = scalar_reg->ss_rel_oid_len;
	bail_require(vars[i].namelen <= VARSTRUCT_oid_size);

	memcpy(vars[i].name, scalar_reg->ss_rel_oid, 
	       sizeof(oid) * scalar_reg->ss_rel_oid_len);
    }

    *ret_offset = offset + num_scalar_regs;

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_reg_table_reg_to_structs(const sn_table *table_reg, VARSTRUCT *vars,
			    uint32 offset, uint32 *ret_offset)
{
    int err = 0;
    uint32 num_field_regs = 0, i = 0;
    sn_field *field_reg = NULL;

    bail_null(table_reg);
    bail_null(vars);
    bail_null(ret_offset);

    err = array_length(table_reg->st_fields, &num_field_regs);
    bail_error(err);

    for (i = offset; i < offset + num_field_regs; ++i) {
	field_reg = array_get_quick(table_reg->st_fields, i - offset);
	bail_null(field_reg);

	bail_require(sizeof(void *) <= sizeof(long));
	vars[i].magic = (long)field_reg;

	vars[i].type = field_reg->sf_snmp_field_type;
	vars[i].acl = RONLY;
	vars[i].findVar = sn_get_handler_columnar;

	err = sn_oid_concat
	    (vars[i].name, VARSTRUCT_oid_size, &(vars[i].namelen), 2,
	     table_reg->st_rel_oid, table_reg->st_rel_oid_len,
	     field_reg->sf_rel_oid, field_reg->sf_rel_oid_len);
	bail_error(err);
    }

    *ret_offset += num_field_regs;

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static void
sn_mib_free(void *elem)
{
    sn_mib *mib = elem;

    if (mib) {
	array_free(&(mib->sm_scalars));
	array_free(&(mib->sm_tables));
	safe_free(mib);
    }
}


/* ------------------------------------------------------------------------- */
static void
sn_scalar_free(void *elem)
{
    sn_scalar **my_elem = (sn_scalar **)&elem;
    safe_free(*my_elem);
}


/* ------------------------------------------------------------------------- */
static void
sn_table_free(void *elem)
{
    sn_table *table = elem;

    if (table) {
	array_free(&(table->st_fields));
	safe_free(table);
    }
}


/* ------------------------------------------------------------------------- */
static void
sn_field_free(void *elem)
{
    sn_field *field = elem;

    if (field) {
	safe_free(field->sf_wildcard_parent);
	safe_free(field);
    }
}


/* ------------------------------------------------------------------------- */
static int 
sn_notif_free(void **key, int32 key_len, void **value)
{
    int err = 0;
    tstring **event_name = NULL;
    sn_notif *notif = NULL;

    /*
     * Don't free the key.  It's part of the sn_notif structure and 
     * probably points to a string constant.
     */

    if (value) {
	notif = *((sn_notif **)value);
	if (notif) {
	    safe_free(notif->sno_vars);
	    safe_free(notif);
	    *value = NULL;
	}
    }

 bail:
    return(err);
}
