/*
 *
 * URL:       $URL$
 * Revision:  $Revision: #1 $
 * Date:      $Date: 2017/11/02 $
 * Author:    $Author: build $
 *
 * (C) Copyright 2002-2012 Riverbed Technology, Inc.
 * All rights reserved.
 *
 */

static const char rcsid[] = "$Id: //depot/prod/cag/R4.4.1x/net-snmp/net-snmp-5.7.1/agent/sn_tables.c#1 $";

#include "common.h"
#include "sn_main.h"
#include "sn_registration.h"
#include "sn_utils.h"
#include "md_client.h"

#define DB_NAME field->sf_table ? field->sf_table->st_mgmt_db_name : NULL


/* ------------------------------------------------------------------------- */
/** Get the value of a field from the first row of a table.
 */
static int
sn_table_get_from_first_row(const sn_field *field,
			    const oid *req_oid, size_t req_oid_len, 
                            tbool exact,
			    oid *ret_oid, size_t *ret_oid_len,
			    void *ret_result, size_t result_buf_size,
			    size_t *ret_result_len);

/* ------------------------------------------------------------------------- */
/** Get the value of a field from the first row of a table having multiple
  indices.
 */
static int
sn_table_get_from_first_row_composite(const sn_field *field,
			    const oid *req_oid, size_t req_oid_len, 
                tbool exact,
			    oid *ret_oid, size_t *ret_oid_len,
			    void *ret_result, size_t result_buf_size, 
			    size_t *ret_result_len);

/* ------------------------------------------------------------------------- */
/** Get the value of a field from the row of a table specified in the 
 * requested OID.
 */
static int
sn_table_get_from_exact_row(const sn_field *field,
			    const oid *req_oid, size_t req_oid_len, 
                            tbool exact,
			    oid *ret_oid, size_t *ret_oid_len,
			    void *ret_result, size_t result_buf_size,
			    size_t *ret_result_len);

/* ------------------------------------------------------------------------- */
/** Get the value of a field from the row of a table specified in the 
 * requested OID in a table having multiple indices.
 */
static int
sn_table_get_from_exact_row_composite(const sn_field *field,
			    const oid *req_oid, size_t req_oid_len, 
                            tbool exact,
			    oid *ret_oid, size_t *ret_oid_len,
			    void *ret_result, size_t result_buf_size,
			    size_t *ret_result_len);

/* ------------------------------------------------------------------------- */
/** Get the value of a field from the row of a table immediately
 * following the one specified in the requested OID.  Note that the
 * requested OID itself need not specify a valid row (unless the index
 * type is sfs_wildcard_sequence); we just want whatever would have
 * come after it had it existed.
 */
static int
sn_table_get_from_next_row(const sn_field *field,
			   const oid *req_oid, size_t req_oid_len, 
                           tbool exact,
			   oid *ret_oid, size_t *ret_oid_len,
			   void *ret_result, size_t result_buf_size,
			   size_t *ret_result_len);

/* ------------------------------------------------------------------------- */
/* Same as sn_table_get_from_next_row, but for tables with multiple indices */

static int
sn_table_get_from_next_row_composite(const sn_field *field,
			   const oid *req_oid, size_t req_oid_len, 
                           tbool exact,
			   oid *ret_oid, size_t *ret_oid_len,
			   void *ret_result, size_t result_buf_size,
			   size_t *ret_result_len);


/* ------------------------------------------------------------------------- */
/* Get the OID for a variable instance to be returned.  inout_oid and
 * inout_oid_len already contain the base OID of the field in
 * question.  This function just appends index values to the OID based
 * on the indices registered in the specified table, and the node name
 * provided which should match the pattern specified in the table
 * registration, with the wildcards filled in with literals.
 */
static int
sn_table_get_field_oid(const sn_table *table, const char *node_name,
		       oid *inout_oid, size_t *inout_oid_len);


/* ------------------------------------------------------------------------- */
/* Get the value of an SNMP variable to be returned from the get
 * handler.
 */
static int
sn_table_get_field_value(const sn_field *field, const char *node_name,
			 void *ret_result, size_t result_buf_size, 
			 size_t *ret_result_len);

/* ------------------------------------------------------------------------- */
/* Get the value of an SNMP variable to be returned from the get
 * handler for tables with multiple indices.
 */
static int
sn_table_get_field_value_composite(const sn_field *field, const char *node_name,
			 void *ret_result, size_t result_buf_size, 
			 size_t *ret_result_len, tstring *new_mgmt_child_name);


/* ------------------------------------------------------------------------- */
/*
 * Compare the requested OID with the one we registered.
 * There are four cases:
 *
 *   (a) The requested one is lower.  That means we must have a 
 *       GETNEXT request (exact == 0), and that we should return
 *       the first element in the column.
 *
 *   (b) The requested one is higher, ours is a prefix of it, and
 *       we have a GET request (exact == true).  We should return
 *       the element specified by the OID provided, if it exists.
 *
 *   (c) The requested one is higher, ours is a prefix of it, and
 *       we have a GETNEXT request (exact == false).  We should try
 *       to find the next element after the OID specified 
 *       (which itself may or may not specify a valid element),
 *       if it exists.
 *
 *   (d) The requested one matches, or the requested one is higher
 *       and ours is not a prefix of it.  This should never happen
 *       and means that either NET-SNMP screwed up or I misunderstood
 *       what it's supposed to do.
 */
int
sn_table_get_field(const sn_field *field,
		   const oid *base_oid, size_t base_oid_len,
		   const oid *req_oid, size_t req_oid_len, 
                   tbool exact,
		   oid *ret_oid, size_t *ret_oid_len,
		   void *ret_result, size_t result_buf_size,
		   size_t *ret_result_len)
{
    int err = 0;
    int diff = 0;
    char buffer[1023];

    *ret_result_len = -1;

    /*
     * We know whatever we return will start with the base OID, so
     * begin with that.  We'll just add whatever we want when we
     * determine the values for the indices.
     */
    memcpy(ret_oid, base_oid, base_oid_len * sizeof(oid));
    *ret_oid_len = base_oid_len;

    diff = snmp_oid_compare(req_oid, req_oid_len, base_oid, base_oid_len);

    /* ........................................................................
     * Case (a): get value from first row 
     */
    if (diff <= 0 && !exact) {

    if (field->sf_table->st_num_indices > 1) {
        err = sn_table_get_from_first_row_composite
	    (field, req_oid, req_oid_len, exact, ret_oid, ret_oid_len,
	     ret_result, result_buf_size, ret_result_len);
    } else {
        err = sn_table_get_from_first_row
            (field, req_oid, req_oid_len, exact, ret_oid, ret_oid_len,
             ret_result, result_buf_size, ret_result_len);
    }
    
	bail_error(err);
    }

    else if ((diff > 0) &&
	     (req_oid_len >= base_oid_len) &&
	     (!memcmp(req_oid, base_oid, base_oid_len))) {

    /* ....................................................................
     * Case (b): get value from exact row specified
     */
	if (exact) {
        if (field->sf_table->st_num_indices > 1) {
            err = sn_table_get_from_exact_row_composite
                (field, req_oid, req_oid_len, exact, ret_oid, ret_oid_len, 
                 ret_result, result_buf_size, ret_result_len);
            bail_error(err);
        } else {
            err = sn_table_get_from_exact_row
                (field, req_oid, req_oid_len, exact, ret_oid, ret_oid_len, 
                 ret_result, result_buf_size, ret_result_len);
            bail_error(err);
        }
	}

    /* ....................................................................
     * Case (c): get value from next row following one specified
     */
	else {

        if (field->sf_table->st_num_indices > 1) {
            err = sn_table_get_from_next_row_composite
                (field, req_oid, req_oid_len, exact, ret_oid, ret_oid_len,
                 ret_result, result_buf_size, ret_result_len);
        } else {
            err = sn_table_get_from_next_row
                (field, req_oid, req_oid_len, exact, ret_oid, ret_oid_len,
                 ret_result, result_buf_size, ret_result_len);
        }
	    bail_error(err);
	}
    }

    /* ........................................................................
     * Case (d): error
     */
    else {
	/* XXX/EMT: could log something more informative */
	bail_require(lc_err_generic_error);
    }

 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
/* There are three steps to handling each of these requests:
 *
 *   1. Find the correct row.  i.e. find the instance of the wildcard(s)
 *      in the management hierarchy that maps to the required row,
 *      if any.  If one cannot be found, stop here.
 *
 *      (a) Get the XXX/EMT
 *
 *   2. Compute the OID to be returned.  This involves taking the base
 *      OID of the field and appending the values for the index or
 *      indices that identify this row, which will depend on the
 *      type of index (wildcard value, child value, or synthetic)
 *
 *   3. Compute the value to be returned.  This process will depend
 *      on the type of the field (wildcard value, child value, or
 *      synthetic).  Note that non-index fields cannot be synthetic,
 *      but that the index is a field like any other and if its
 *      value is requested it will go through the same path as the
 *      other fields.
 *
 * Step 1 is a different process for each of the three functions
 * below: to get the first, next, or an exact row.  Steps 2 and 3 are
 * the same for all of those cases.
 *
 * Note: for version 1.0, we are only supporting one index, and
 * indices must be synthetic.  This simplifies the code below a lot.
 */

/* ------------------------------------------------------------------------- */
static int
sn_table_get_from_first_row(const sn_field *field,
			    const oid *req_oid, size_t req_oid_len, 
                            tbool exact,
			    oid *ret_oid, size_t *ret_oid_len,
			    void *ret_result, size_t result_buf_size, 
			    size_t *ret_result_len)
{
    int err = 0;
    const tstring *node_name = NULL;
    bn_binding *binding = NULL;
    sn_field *idx = NULL;
    sn_table *table = NULL;
    tstring *idx_seq_str = NULL;

    table = field->sf_table;
    bail_null(table);
    idx = (sn_field *) array_get_quick(table->st_fields, 0);
    bail_null(idx);

    /*
     * For getting the first row, the index type doesn't matter.
     */
    err = mdc_get_nth_binding_child(NULL, NULL, true, &binding, 
                                    idx->sf_mgmt_type,
				    0, idx->sf_wildcard_parent, DB_NAME);
    bail_error(err);

    if (binding == NULL) {
	goto bail;
    }

    /*
     * Normally the index of the first row would be 1 (we set it to 0
     * here because it gets incremented elsewhere).  But if the index
     * is derived directly from the wildcard, we have to take its
     * value instead.
     */
    if (idx->sf_source == sfs_wildcard_value) {
	err = bn_binding_get_name_last_part(binding, &idx_seq_str);
	bail_error(err);
	idx->sf_instance_seq = strtol(ts_str(idx_seq_str), NULL, 10);
    }
    else {
	idx->sf_instance_seq = 1;
    }

    err = bn_binding_get_name(binding, &node_name);
    bail_error(err);

    err = sn_table_get_field_oid
	(field->sf_table, ts_str(node_name), ret_oid, ret_oid_len);
    bail_error(err);

    err = sn_table_get_field_value
	(field, ts_str(node_name), ret_result, result_buf_size, 
	 ret_result_len);
    bail_error(err);
 bail:
    ts_free(&idx_seq_str);
    bn_binding_free(&binding);
    return(err);
}
/* ------------------------------------------------------------------------- */
static int
sn_table_get_from_first_row_composite(const sn_field *field,
			    const oid *req_oid, size_t req_oid_len, 
                tbool exact,
			    oid *ret_oid, size_t *ret_oid_len,
			    void *ret_result, size_t result_buf_size, 
			    size_t *ret_result_len)
{
    int err = 0, last_wildcard = 0;
    const tstring *node_name = NULL;
    const char *value_str = NULL;
    bn_binding *binding = NULL;
    sn_field *idx = NULL;
    sn_table *table = NULL;
    tstring *idx_seq_str = NULL;
    tstring *wildcard_parent = NULL, *new_child_node = NULL,
            *modified_child = NULL;
    tstr_array *name_parts = NULL, *child_parts = NULL;
    int i = 0, num_fields = 0, num_parts = 0, part_index = 0, rem_parts = 0;

    table = field->sf_table;
    bail_null(table);
    err = ts_new_str(&wildcard_parent, field->sf_wildcard_parent);
    bail_error(err);
    err = array_length(table->st_fields, &num_fields);
    bail_error(err);
    err = ts_new_str(&new_child_node, "");

    /* for each of the index nodes 
       do
    */
    for ( i = 0 ; i < num_fields ; ++i) {

        idx = (sn_field *) array_get_quick(table->st_fields, i);
        bail_null(idx);

    /* We don't care about non-index nodes*/
        if (!idx->sf_index) {
            continue;
        }

    /*
       We need to attach all the nodes upto "*" to the wildcard parent, 
       in order to handle the case where different indices are served by 
       wildcards that are not at the same level in the mgmt tree.

       We are checking for i > 0 , because for i = 0 the framework is anyway
       able to calculate the wildcard_parent correctly. It is only from second
       wildcard onwards that the framework fails to calculate the wildcard 
       parent.
     */

        if (i) {
            err = bn_binding_name_to_parts(idx->sf_snmp_field_name, 
                    &name_parts, NULL);
            bail_error(err);
            while ((value_str = tstr_array_get_str_quick
                        (name_parts, part_index++))) {
                if (!strcmp(value_str, "*")) {
                    rem_parts = part_index + 1;
                    while((value_str = tstr_array_get_str_quick
                                (name_parts, rem_parts++))) {
                        if (strcmp(value_str, "*")) {
                            err = ts_append_str(new_child_node, "/");
                            bail_error(err);
                            err = ts_append_str(new_child_node, value_str);
                            bail_error(err);
                        }
                        else {
                            break;
                        }
                    }
                    break;
                } else {
                    err = ts_append_str(wildcard_parent, "/");
                    bail_error(err);
                    err = ts_append_str(wildcard_parent, value_str);
                    bail_error(err);
                }
            }
            tstr_array_free(&name_parts);

        }
        /*
           Here, we get the first child by passing zero to the sixth argument 
         */

        err = mdc_get_nth_binding_child(NULL, NULL, true, &binding, 
                idx->sf_mgmt_type,
                0, ts_str(wildcard_parent), DB_NAME); 
        bail_error(err);

        if (binding == NULL) {
            goto bail;
        }

        /*
         * Normally the index of the first row would be 1 (we set it to 0
         * here because it gets incremented elsewhere).  But if the index
         * is derived directly from the wildcard, we have to take its
         * value instead.
         *
         * This currently works for numeric indices. At some point we need 
         * to make this work for other types. For example "1000" is a valid 
         * index, whereas "riverbed" or "10.0.0.1" is not.
         */

        if (idx->sf_source == sfs_wildcard_value) {
            err = bn_binding_get_name_last_part(binding, &idx_seq_str);
            bail_error(err);
            idx->sf_instance_seq = strtol(ts_str(idx_seq_str), NULL, 10);
            ts_free(&idx_seq_str);
        }
        else {
            idx->sf_instance_seq = 1;
        }

        err = bn_binding_get_name(binding, &node_name);
        bail_error(err);

        /*
           From the second index onwards we need to manually expand the wildcard
           parent to replace the * for the prev indices with their corresponding
           instance values.
         */

        ts_free(&wildcard_parent);
        err = ts_new_str(&wildcard_parent, ts_str(node_name));
        bail_error(err);
    }

    err = sn_table_get_field_oid
	(field->sf_table, ts_str(node_name), ret_oid, ret_oid_len);
    bail_error(err);

    if (field->sf_table->st_num_indices > 1) {
        if (field->sf_source == sfs_child_value) {
            err = bn_binding_name_to_parts
                (field->sf_mgmt_node_child_name, &child_parts, NULL);
            bail_error(err);

            part_index = 0;
            while ((value_str = tstr_array_get_str_quick
                    (child_parts, ++part_index))) {
                if (!strcmp(value_str, "*")) {
                    last_wildcard = part_index;
                }
            }
            for (i = 0 ; i < last_wildcard + 1; ++i) {
                err = tstr_array_delete(child_parts, 0);
                bail_error(err);
            }

            err = bn_binding_name_parts_to_name_tstr
                (child_parts, true, &modified_child);
            bail_error(err);
            tstr_array_free(&child_parts);
        }
        err = sn_table_get_field_value_composite
            (field, ts_str(node_name), ret_result, result_buf_size, 
             ret_result_len, modified_child);
    } else {
        err = sn_table_get_field_value
            (field, ts_str(node_name), ret_result, result_buf_size, 
             ret_result_len);
    }
    bail_error(err);

 bail:
    ts_free(&idx_seq_str);
    ts_free(&new_child_node);
    ts_free(&wildcard_parent);
    bn_binding_free(&binding);
    tstr_array_free(&name_parts);
    tstr_array_free(&child_parts);
    return(err);
}

/* ------------------------------------------------------------------------- */

static int
sn_table_get_from_exact_row(const sn_field *field,
			    const oid *req_oid, size_t req_oid_len, 
                            tbool exact,
			    oid *ret_oid, size_t *ret_oid_len,
			    void *ret_result, size_t result_buf_size, 
			    size_t *ret_result_len)
{
    int err = 0;
    sn_table *table = NULL;
    sn_field *idx = NULL;
    const tstring *node_name = NULL;
    uint32 req_seq = 0;
    bn_binding *binding = NULL;
    char *bname = NULL;
    
    table = field->sf_table;
    bail_null(table);
    idx = (sn_field *) array_get_quick(table->st_fields, 0);
    bail_null(idx);

    switch (idx->sf_source) {
    case sfs_wildcard_sequence:
	req_seq = req_oid[req_oid_len - 1];

	/*
	 * Subtract 1 from req_seq because SNMP starts counting from 1,
	 * while we start counting from 0.
	 */ 
	err = mdc_get_nth_binding_child(NULL, NULL, true, &binding,
                                        idx->sf_mgmt_type,
					req_seq - 1, idx->sf_wildcard_parent, 
                                        DB_NAME);
	bail_error(err);
	
	if (binding == NULL) {
	    goto bail;
	}
	
	idx->sf_instance_seq = req_seq;
	break;

    case sfs_wildcard_value:
	req_seq = req_oid[req_oid_len - 1];

	bname = smprintf("%s/%d", idx->sf_wildcard_parent, req_seq);
	bail_null(bname);

	err = mdc_get_binding(NULL, NULL, true, &binding, bname, DB_NAME);
	bail_error(err);

	if (binding == NULL) {
	    goto bail;
	}
	idx->sf_instance_seq = req_seq;
	break;

    case sfs_child_value:
	bail_force(lc_err_nyi);

    default:
	bail_force(lc_err_unexpected_case);

    }

    err = bn_binding_get_name(binding, &node_name);
    bail_error(err);

    err = sn_table_get_field_oid
	(field->sf_table, ts_str(node_name), ret_oid, ret_oid_len);
    bail_error(err);

    err = sn_table_get_field_value
	(field, ts_str(node_name), ret_result, result_buf_size,
	 ret_result_len);
    bail_error(err);
 bail:
    safe_free(bname);
    bn_binding_free(&binding);
    return(err);
}

/* ------------------------------------------------------------------------- */
static int
sn_table_get_from_exact_row_composite(const sn_field *field,
			    const oid *req_oid, size_t req_oid_len, 
                            tbool exact,
			    oid *ret_oid, size_t *ret_oid_len,
			    void *ret_result, size_t result_buf_size, 
			    size_t *ret_result_len)
{
    int err = 0, i = 0, num_fields = 0, num_indices = 0, nth_index = 0, 
        part_index = 0, rem_parts = 0, last_wildcard = 0;
    sn_table *table = NULL;
    sn_field *idx = NULL;
    const tstring *node_name = NULL;
    uint32 req_seq = 0;
    bn_binding *binding = NULL;
    char buffer[1023];
    uint32 indices[64];
    tstr_array *name_parts = NULL;
    const char *value_str = NULL;
    tstring *idx_seq_str = NULL;
    tstring *wildcard_parent = NULL, *new_child_node = NULL;
    char *wildcard_value = NULL;
    tstr_array *child_parts = NULL;
  
    table = field->sf_table;
    bail_null(table);
    idx = (sn_field *) array_get_quick(table->st_fields, 0);
    bail_null(idx);
 
    err = ts_new_str(&wildcard_parent, field->sf_wildcard_parent);
    bail_error(err);

    err = array_length(table->st_fields, &num_fields);
    bail_error(err);

    num_indices = table->st_num_indices;
    err = ts_new_str(&new_child_node, "");

    /*   Just trying to store the copied indices from left->to->right, even
         though we are retrieving them from right->to->left.
         -1 is to account for arrays starting from zero 
     */
    
    for (i = 0 ; i < num_indices ; ++i) {
        indices[num_indices - i - 1] = req_oid[req_oid_len - i - 1];
    }

    for ( i = 0 ; i < num_fields ; ++i) {

        idx = (sn_field *) array_get_quick(table->st_fields, i);
        bail_null(idx);

        /* We don't care about non-index nodes */
       
        if (!idx->sf_index) {
            continue;
        } 
        bn_binding_free(&binding);
    
     /* 
       We need to attach all the nodes upto "*" to the wildcard parent, 
       in order to handle the case where different indices are served by 
       wildcards that are not at the same level in the mgmt tree.
     */
     
        if (i) {
            err = bn_binding_name_to_parts(idx->sf_snmp_field_name, 
                    &name_parts, NULL);
            bail_error(err);
            while ((value_str = tstr_array_get_str_quick(name_parts
                            , part_index++))) {
                if (!strcmp(value_str, "*")) {
                    rem_parts = part_index + 1;
                    while ((value_str = tstr_array_get_str_quick
                                (name_parts, rem_parts++))) {
                        if (strcmp(value_str, "*")) {
                            err = ts_append_str(new_child_node, "/");
                            bail_error(err);
                            err = ts_append_str(new_child_node, value_str);
                            bail_error(err);
                        }
                        else {
                            break;
                        }
                    }
                    break;
                } else {
                    err = ts_append_str(wildcard_parent, "/");
                    bail_error(err);
                    err = ts_append_str(wildcard_parent, value_str);
                    bail_error(err);
                }
            }
            tstr_array_free(&name_parts);
        }

        if (idx->sf_source == sfs_wildcard_sequence) {
            err = mdc_get_nth_binding_child(NULL, NULL, true, &binding, 
                    idx->sf_mgmt_type,
                    indices[nth_index] - 1, ts_str(wildcard_parent), DB_NAME); 
        } else {
            wildcard_value = smprintf("%d", indices[nth_index]);
            err = ts_append_str(wildcard_parent, "/");
            bail_error(err);
            err = ts_append_str(wildcard_parent, wildcard_value);
            bail_error(err);
            safe_free(wildcard_value);
            err = mdc_get_binding(NULL, NULL, true, &binding, 
                    ts_str(wildcard_parent), DB_NAME);
        }
        bail_error(err);
        bail_null(binding);

    
     /* Normally the index of the first row would be 1 (we set it to 0
     * here because it gets incremented elsewhere).  But if the index
     * is derived directly from the wildcard, we have to take its
     * value instead.
     */
     
        if (idx->sf_source == sfs_wildcard_value) {
            err = bn_binding_get_name_last_part(binding, &idx_seq_str);
            bail_error(err);
            idx->sf_instance_seq = strtol(ts_str(idx_seq_str), NULL, 10);
            ts_free(&idx_seq_str);
        }
        else {
            idx->sf_instance_seq = indices[nth_index];
        }

        err = bn_binding_get_name(binding, &node_name);
        bail_error(err);

        /* 
           From the second index onwards we need to manually expand the wildcard
           parent to replace the * for the prev indices with their corresponding
           instance values.
         */

        ts_free(&wildcard_parent);
        err = ts_new_str(&wildcard_parent, ts_str(node_name));
        bail_error(err);
        ++nth_index;
    }
    
    err = sn_table_get_field_oid
	(field->sf_table, ts_str(node_name), ret_oid, ret_oid_len);
    bail_error(err);

    if (field->sf_table->st_num_indices > 1) {
        if (field->sf_source == sfs_child_value) {
            err = bn_binding_name_to_parts
                (field->sf_mgmt_node_child_name, &child_parts, NULL);
            bail_error(err);

            part_index = 0;
            while ((value_str = tstr_array_get_str_quick
                    (child_parts, ++part_index))) {
                if (!strcmp(value_str, "*")) {
                    last_wildcard = part_index;
                }
            }
            for (i = 0 ; i < last_wildcard + 1; ++i) {
                err = tstr_array_delete(child_parts, 0);
                bail_error(err);
            }
            err = bn_binding_name_parts_to_name_tstr
                (child_parts, true, &new_child_node);
            bail_error(err);
            tstr_array_free(&child_parts);
        }

        err = sn_table_get_field_value_composite
            (field, ts_str(node_name), ret_result, result_buf_size,
             ret_result_len, new_child_node);
    } 
    else {
        err = sn_table_get_field_value
            (field, ts_str(node_name), ret_result, result_buf_size,
             ret_result_len);
    }

    bail_error(err);
    
 bail:
    bn_binding_free(&binding);
    tstr_array_free(&name_parts);
    tstr_array_free(&child_parts);
    ts_free(&idx_seq_str);
    ts_free(&new_child_node);
    ts_free(&wildcard_parent);
    
    return(err);
}

/* ------------------------------------------------------------------------- */
static int
sn_table_get_from_next_row(const sn_field *field,
			   const oid *req_oid, size_t req_oid_len, 
                           tbool exact,
			   oid *ret_oid, size_t *ret_oid_len,
			   void *ret_result, size_t result_buf_size, 
			   size_t *ret_result_len)
{
    int err = 0;
    sn_table *table = NULL;
    sn_field *idx = NULL;
    bn_binding *binding = NULL;
    const tstring *node_name = NULL;
    uint32 req_seq = 0;
    char *req_seq_str = NULL;
    tstring *last_part = NULL;
    
    table = field->sf_table;
    bail_null(table);
    idx = (sn_field *) array_get_quick(table->st_fields, 0);
    bail_null(idx);

    switch (idx->sf_source) {

    case sfs_wildcard_sequence:
	req_seq = req_oid[req_oid_len - 1];

	/*
	 * Subtract 1 from req_seq because SNMP starts counting from 1,
	 * while we start counting from 0.
	 */ 
	err = mdc_get_nth_binding_child(NULL, NULL, true, &binding,
                                        idx->sf_mgmt_type,
					req_seq + 1 - 1,
					idx->sf_wildcard_parent, DB_NAME);
	bail_error(err);
	
	if (binding == NULL) {
	    goto bail;
	}
	
	idx->sf_instance_seq = req_seq + 1;
	break;

    case sfs_wildcard_value:   
	req_seq = req_oid[req_oid_len - 1];
	req_seq_str = smprintf("%d", req_seq);
	bail_null(req_seq_str);

	err = mdc_get_next_binding_child
	    (NULL, NULL, true, &binding, idx->sf_mgmt_type, 
             idx->sf_wildcard_parent,
	     req_seq_str, DB_NAME);
	bail_error_msg(err, "Couldn't get next for item %s", req_seq_str);
	
	if (binding == NULL) {
	    goto bail;
	}

	err = bn_binding_get_name_last_part(binding, &last_part);
	bail_error(err);
    
	idx->sf_instance_seq = strtoul(ts_str(last_part), NULL, 10); 

	break;

    case sfs_child_value:
	bail_force(lc_err_nyi);

    default:
	bail_force(lc_err_unexpected_case);

    }

    err = bn_binding_get_name(binding, &node_name);
    bail_error(err);

    err = sn_table_get_field_oid
	(field->sf_table, ts_str(node_name), ret_oid, ret_oid_len);
    bail_error(err);

    err = sn_table_get_field_value
	(field, ts_str(node_name), ret_result, result_buf_size,
	 ret_result_len);
    bail_error(err);

 bail:
    ts_free(&last_part);
    safe_free(req_seq_str);
    bn_binding_free(&binding);
    return(err);
}

    static int
sn_table_get_from_next_row_composite(const sn_field *field,
			   const oid *req_oid, size_t req_oid_len, 
                           tbool exact,
			   oid *ret_oid, size_t *ret_oid_len,
			   void *ret_result, size_t result_buf_size, 
			   size_t *ret_result_len)
{
    int err = 0, i = 0, num_fields = 0, num_indices = 0, nth_index = 0, 
        part_index = 0, rem_parts = 0, last_wildcard = 0, getnext_index =0,
        current_index = 0;
    sn_table *table = NULL;
    sn_field *idx = NULL;
    const tstring *node_name = NULL;
    uint32 req_seq = 0;
    bn_binding *binding = NULL;
    char buffer[1023];
    uint32 indices[64];
    tbool  tried_next_index[64];
    tstr_array *name_parts  = NULL;
    const char *value_str = NULL;
    tstring *idx_seq_str = NULL, *wildcard_parent = NULL, 
            *new_child_node = NULL, *modified_child = NULL;
    char *wildcard_value = NULL;
    tstr_array *child_parts = NULL;
    tbool node_not_found = true;

    table = field->sf_table;
    bail_null(table);
    err = array_length(table->st_fields, &num_fields);
    bail_error(err);
    err = ts_new_str(&new_child_node, "");
    bail_error(err);
    num_indices = table->st_num_indices;
    getnext_index = num_indices;

    /* Store the indices from left->to->right.
     -1 is to account for arrays starting from zero 
     */
    for (i = 0 ; i < num_indices ; ++i) {
        indices[num_indices - i - 1] = req_oid[req_oid_len - i - 1];
    }
    for (i = 0 ; i < num_indices ; ++i) {
        tried_next_index[i] = false;
    }

    while (node_not_found) {
        if (!getnext_index) {
            goto bail;
        }
        err = ts_new_str(&wildcard_parent, field->sf_wildcard_parent);
        bail_error(err);
        err = ts_new_str(&new_child_node, "");
        bail_error(err);
        nth_index = current_index = 0;

        for ( i = 0 ; i < num_fields && node_not_found; ++i) {
            bn_binding_free(&binding);
            idx = (sn_field *) array_get_quick(table->st_fields, i);
            bail_null(idx);

            /* We don't care about non-index nodes  */
            if (!idx->sf_index) {
                continue;
            } 

            ++current_index;

            /*     Wildcard parent for the first wildcard is returned correctly
                   by framework. We only need to generate the wildcard parent, 
                   starting from the second wildcard
             */
            
            if (i) {
                err = bn_binding_name_to_parts(idx->sf_snmp_field_name, 
                        &name_parts, NULL);
                bail_error(err);
                part_index = 0;

                /*We need to attach all the nodes upto "*" to the 
                  wildcard parent 
                */
                while ((value_str = tstr_array_get_str_quick(name_parts
                            , part_index++))) {
                    if (!strcmp(value_str, "*")) {
                        rem_parts = part_index + 1;
                        while((value_str = tstr_array_get_str_quick
                                (name_parts, rem_parts++))) {
                            if (strcmp(value_str, "*")) {
                                err = ts_append_str(new_child_node, "/");
                                bail_error(err);
                                err = ts_append_str(new_child_node, value_str);
                                bail_error(err);
                            }
                            else {
                                break;
                            }
                        }
                        break;
                    } else {
                        err = ts_append_str(wildcard_parent, "/");
                        bail_error(err);
                        err = ts_append_str(wildcard_parent, value_str);
                        bail_error(err);
                    }
                }
                tstr_array_free(&name_parts);
            }

            if (idx->sf_source == sfs_wildcard_sequence) {
                if (current_index == getnext_index) {
                    err = mdc_get_nth_binding_child(NULL, NULL, true, &binding,
                            idx->sf_mgmt_type, indices[nth_index] , 
                            ts_str(wildcard_parent), DB_NAME);
                    bail_error(err);
                    if (!binding) {
                        --getnext_index;
                        if (current_index) {
                            tried_next_index[current_index] = false;
                        }
                        indices[nth_index] = 0;
                        break;
                    } else if (current_index == num_indices) {
                        node_not_found = false;
                    }
                } else {
                    err = mdc_get_nth_binding_child(NULL, NULL, true, &binding, 
                            idx->sf_mgmt_type, indices[nth_index] - 1, 
                            ts_str(wildcard_parent), DB_NAME);
                    bail_error(err);
                    bail_null(binding);
                } 
            } else {
                /* If the index is of type wildcard_value */
                wildcard_value = smprintf("%d", indices[nth_index]);
                if (current_index == getnext_index) {
                    err = mdc_get_next_binding_child(NULL, NULL, true
                            , &binding, idx->sf_mgmt_type,                     
                            ts_str(wildcard_parent), wildcard_value, 
                            DB_NAME);
                    if (err != 14004) {
                        bail_error(err);
                    }
                    safe_free(wildcard_value);

                    if (!binding) {
                        --getnext_index;
                        break;
                    } else {
                        node_not_found = false;
                    }
                } else {
                    err = mdc_get_nth_binding_child(NULL, NULL, true
                            , &binding, 
                            idx->sf_mgmt_type,
                            0 , ts_str(wildcard_parent), DB_NAME);
                    bail_error(err);
                    bail_null(binding);
                }
            }
            
            /* Compute the index to be appended to the oid to form the 
               instance id.
             */ 

            if (idx->sf_source == sfs_wildcard_value) {
                err = bn_binding_get_name_last_part(binding, &idx_seq_str);
                bail_error(err);
               idx->sf_instance_seq = strtol(ts_str(idx_seq_str), NULL, 10);
                ts_free(&idx_seq_str);

            }
            else {
                /* If the index is wildcard_sequence just use the next higher
                   integer to the one in the request oid. 
                 */

                idx->sf_instance_seq = indices[nth_index];
                if (current_index == getnext_index) {
                    ++idx->sf_instance_seq;
                    if (current_index != num_indices && 
                            !tried_next_index[current_index]) {
                        ++getnext_index;
                        tried_next_index[current_index] = true;
                    }
                }
            }
            err = bn_binding_get_name(binding, &node_name);
            bail_error(err);
            ts_free(&wildcard_parent);
            err = ts_new_str(&wildcard_parent, ts_str(node_name));
            bail_error(err);
            ++nth_index;
        }
        ts_free(&wildcard_parent);
        ts_free(&new_child_node);
    }
    
    err = sn_table_get_field_oid
	(field->sf_table, ts_str(node_name), ret_oid, ret_oid_len);
    bail_error(err);

    if (field->sf_table->st_num_indices > 1) {
        if (field->sf_source == sfs_child_value) {
            err = bn_binding_name_to_parts
                (field->sf_mgmt_node_child_name, &child_parts, NULL);
            bail_error(err);

            part_index = 0;
            while ((value_str = tstr_array_get_str_quick
                    (child_parts, ++part_index))) {
                if (!strcmp(value_str, "*")) {
                    last_wildcard = part_index;
                }
            }
            for (i = 0 ; i < last_wildcard + 1; ++i) {
                err = tstr_array_delete(child_parts, 0);
                bail_error(err);
            }
            err = bn_binding_name_parts_to_name_tstr
                (child_parts, true, &modified_child);
            bail_error(err);
            tstr_array_free(&child_parts);
        }
        err = sn_table_get_field_value_composite
            (field, ts_str(node_name), ret_result, result_buf_size,
             ret_result_len, modified_child);
        bail_error(err);
    } 
    else {
        err = sn_table_get_field_value
            (field, ts_str(node_name), ret_result, result_buf_size,
             ret_result_len);
        bail_error(err);
    }

 bail:
    bn_binding_free(&binding);
    ts_free(&idx_seq_str);
    ts_free(&wildcard_parent);
    tstr_array_free(&name_parts);
    tstr_array_free(&child_parts);
    safe_free(wildcard_value); 
    return(err);
}

/* ------------------------------------------------------------------------- */
static int
sn_table_get_field_oid(const sn_table *table, const char *node_name,
		       oid *inout_oid, size_t *inout_oid_len)
{
    int err = 0, i = 0, num_fields = 0;
    sn_field *idx = NULL;
    char buffer[1023];

    /*
     * Currently only a single index, and only synthetic indices, are
     * supported, so this is pretty simple.  Take the sequence number
     * out of the index record and tack it onto the end of the OID
     * (after adding 1, since indices count from 0 from SNMP fields
     * count from 1).
     */
    err = array_length(table->st_fields, &num_fields);
    bail_error(err);

    for ( i = 0 ; i < num_fields ; ++i) {

        idx = (sn_field *) array_get_quick(table->st_fields, i);
        bail_null(idx);

        if (!idx->sf_index) { 
            continue;
        } 
    
        inout_oid[(*inout_oid_len)++] = idx->sf_instance_seq;
    }
    
 bail:
    return(err);
}


/* ------------------------------------------------------------------------- */
static int
sn_table_get_field_value(const sn_field *field, const char *node_name,
			 void *ret_result, size_t result_buf_size, 
			 size_t *ret_result_len)
{
    int err = 0;
    char *child_name = NULL;
    tstr_array *name_parts = NULL;
    const char *value_str = NULL;
    bn_binding *value_binding = NULL;
    uint32 idx = 0;

    switch (field->sf_source) {

    case sfs_child_value:
	child_name = smprintf("%s/%s", node_name,
			      field->sf_mgmt_node_child_name);
	bail_null(child_name);
	err = sn_get_snmp_value_by_name
	    (child_name, DB_NAME, field->sf_mgmt_type, ret_result, result_buf_size, 
	     ret_result_len, NULL);

	bail_error(err);
	break;

    case sfs_wildcard_value:
	err = bn_binding_name_to_parts(node_name, &name_parts, NULL);
	bail_error(err);
	value_str = tstr_array_get_str_quick
	    (name_parts, field->sf_wildcard_name_index);
	bail_null(value_str);
	err = bn_binding_new_str(&value_binding, "dummy", ba_value,
				 field->sf_mgmt_type, 0, value_str);
	bail_error_null(err, value_binding);
	err = sn_binding_to_snmp_result(value_binding, field->sf_mgmt_type,
					ret_result, result_buf_size, 
					ret_result_len, NULL);
	bail_error(err);
	break;

    case sfs_wildcard_sequence:
	bail_require(field->sf_index);
	idx = field->sf_instance_seq;
	err = bn_binding_new_uint32(&value_binding, "dummy", ba_value, 0, idx);
	bail_error(err);
	err = sn_binding_to_snmp_result(value_binding, field->sf_mgmt_type,
					ret_result, result_buf_size, 
					ret_result_len, NULL);
	bail_error(err);
	break;

    default:
	bail_force(lc_err_unexpected_case);
    }

 bail:
    bn_binding_free(&value_binding);
    tstr_array_free(&name_parts);
    safe_free(child_name);
    return(err);
}

/* ------------------------------------------------------------------------- */
static int
sn_table_get_field_value_composite(const sn_field *field, const char *node_name,
			 void *ret_result, size_t result_buf_size, 
			 size_t *ret_result_len, tstring *new_mgmt_child_name)
{
    int err = 0;
    char *child_name = NULL;
    tstr_array *name_parts = NULL;
    const char *value_str = NULL;
    bn_binding *value_binding = NULL;
    uint32 idx = 0;

    switch (field->sf_source) {
    case sfs_child_value:
    if (new_mgmt_child_name) {
        child_name = smprintf("%s/%s", node_name, ts_str(new_mgmt_child_name) + 1);
    } else {
        child_name = smprintf("%s/%s", node_name, field->sf_mgmt_node_child_name );
    }

	bail_null(child_name);
	err = sn_get_snmp_value_by_name
	    (child_name, DB_NAME, field->sf_mgmt_type, ret_result, result_buf_size, 
	     ret_result_len, NULL);
	bail_error(err);
	break;

    case sfs_wildcard_value:
    bail_null(node_name);
	err = bn_binding_name_to_parts(node_name, &name_parts, NULL);
	bail_error(err);
	value_str = tstr_array_get_str_quick
	    (name_parts, field->sf_wildcard_part_number);
	bail_null(value_str);
	err = bn_binding_new_str(&value_binding, "dummy", ba_value,
				 field->sf_mgmt_type, 0, value_str);
	bail_error_null(err, value_binding);
	err = sn_binding_to_snmp_result(value_binding, field->sf_mgmt_type,
					ret_result, result_buf_size, 
					ret_result_len, NULL);
	bail_error(err);
	break;

    case sfs_wildcard_sequence:
	bail_require(field->sf_index);
	idx = field->sf_instance_seq;
	err = bn_binding_new_uint32(&value_binding, "dummy", ba_value, 0, idx);
	bail_error(err);
	err = sn_binding_to_snmp_result(value_binding, field->sf_mgmt_type,
					ret_result, result_buf_size, 
					ret_result_len, NULL);
	bail_error(err);
	break;

    default:
	bail_force(lc_err_unexpected_case);
    }

 bail:
    bn_binding_free(&value_binding);
    tstr_array_free(&name_parts);
    safe_free(child_name);
    return(err);
}
